@@ -7,35 +7,41 @@
 {-# LANGUAGE TupleSections     #-}
 module Main where
 
-import           Control.Arrow           (second)
-import           Control.Monad           (when)
-import           Control.Monad.Logger    (LogLevel (..), filterLogger, logInfoS,
-                                          runStderrLoggingT)
-import           Data.Char               (toLower)
-import           Data.Default            (Default (..))
-import           Data.List               (intercalate)
-import           Data.Maybe              (fromMaybe)
-import           Data.String.Conversions (cs)
-import           Haskoin                 (Network (..), allNets, bch,
-                                          bchRegTest, bchTest, btc, btcRegTest,
-                                          btcTest, eitherToMaybe)
-import           Haskoin.Node            (withConnection)
-import           Haskoin.Store           (StoreConfig (..), WebConfig (..),
-                                          WebLimits (..), WebTimeouts (..),
-                                          runWeb, withStore)
-import           Options.Applicative     (Parser, auto, eitherReader,
-                                          execParser, flag, fullDesc, header,
-                                          help, helper, info, long, many,
-                                          metavar, option, progDesc, short,
-                                          showDefault, strOption, switch, value)
-import           System.Exit             (exitSuccess)
-import           System.FilePath         ((</>))
-import           System.IO.Unsafe        (unsafePerformIO)
-import           Text.Read               (readMaybe)
-import           UnliftIO                (MonadIO)
-import           UnliftIO.Directory      (createDirectoryIfMissing,
-                                          getAppUserDataDirectory)
-import           UnliftIO.Environment    (lookupEnv)
+import           Control.Applicative       ((<|>))
+import           Control.Arrow             (second)
+import           Control.Monad             (when)
+import           Control.Monad.Logger      (LogLevel (..), filterLogger,
+                                            logInfoS, runStderrLoggingT)
+import           Control.Monad.Trans.Maybe (MaybeT (..), runMaybeT)
+import           Data.Char                 (toLower)
+import           Data.Default              (Default (..))
+import           Data.List                 (intercalate)
+import           Data.Maybe                (fromMaybe)
+import           Data.String.Conversions   (cs)
+import qualified Data.Text                 as T
+import           Data.Word                 (Word32)
+import           Haskoin                   (Network (..), allNets, bch,
+                                            bchRegTest, bchTest, btc,
+                                            btcRegTest, btcTest, eitherToMaybe)
+import           Haskoin.Node              (withConnection)
+import           Haskoin.Store             (StoreConfig (..), WebConfig (..),
+                                            WebLimits (..), WebTimeouts (..),
+                                            runWeb, withStore)
+import           Haskoin.Store.Stats       (withStats)
+import           Options.Applicative       (Parser, auto, eitherReader,
+                                            execParser, flag, fullDesc, header,
+                                            help, helper, info, long, many,
+                                            metavar, option, progDesc, short,
+                                            showDefault, strOption, switch,
+                                            value)
+import           System.Exit               (exitSuccess)
+import           System.FilePath           ((</>))
+import           System.IO.Unsafe          (unsafePerformIO)
+import           Text.Read                 (readMaybe)
+import           UnliftIO                  (MonadIO)
+import           UnliftIO.Directory        (createDirectoryIfMissing,
+                                            getAppUserDataDirectory)
+import           UnliftIO.Environment      (lookupEnv)
 
 version :: String
 #ifdef CURRENT_PACKAGE_VERSION
@@ -45,47 +51,67 @@ version = "Unavailable"
 #endif
 
 data Config = Config
-    { configDir         :: !FilePath
-    , configHost        :: !String
-    , configPort        :: !Int
-    , configNetwork     :: !Network
-    , configDiscover    :: !Bool
-    , configPeers       :: ![(String, Maybe Int)]
-    , configVersion     :: !Bool
-    , configDebug       :: !Bool
-    , configReqLog      :: !Bool
-    , configWebLimits   :: !WebLimits
-    , configWebTimeouts :: !WebTimeouts
-    , configRedis       :: !Bool
-    , configRedisURL    :: !String
-    , configRedisMin    :: !Int
-    , configRedisMax    :: !Integer
-    , configWipeMempool :: !Bool
-    , configPeerTimeout :: !Int
-    , configPeerMaxLife :: !Int
-    , configMaxPeers    :: !Int
+    { configDir             :: !FilePath
+    , configHost            :: !String
+    , configPort            :: !Int
+    , configNetwork         :: !Network
+    , configAsert           :: !Word32
+    , configDiscover        :: !Bool
+    , configPeers           :: ![(String, Maybe Int)]
+    , configVersion         :: !Bool
+    , configDebug           :: !Bool
+    , configMaxPending      :: !Int
+    , configWebLimits       :: !WebLimits
+    , configWebTimeouts     :: !WebTimeouts
+    , configRedis           :: !Bool
+    , configRedisURL        :: !String
+    , configRedisMin        :: !Int
+    , configRedisMax        :: !Integer
+    , configWipeMempool     :: !Bool
+    , configNoMempool       :: !Bool
+    , configPeerTimeout     :: !Int
+    , configPeerMaxLife     :: !Int
+    , configMaxPeers        :: !Int
+    , configMaxDiff         :: !Int
+    , configCacheRefresh    :: !Int
+    , configCacheRetryDelay :: !Int
+    , configNumTxId         :: !Bool
+    , configStatsd          :: !Bool
+    , configStatsdHost      :: !String
+    , configStatsdPort      :: !Int
+    , configStatsdPrefix    :: !String
     }
 
 instance Default Config where
-    def = Config { configDir         = defDirectory
-                 , configHost        = defHost
-                 , configPort        = defPort
-                 , configNetwork     = defNetwork
-                 , configDiscover    = defDiscover
-                 , configPeers       = defPeers
-                 , configVersion     = False
-                 , configDebug       = defDebug
-                 , configReqLog      = defReqLog
-                 , configWebLimits   = defWebLimits
-                 , configWebTimeouts = defWebTimeouts
-                 , configRedis       = defRedis
-                 , configRedisURL    = defRedisURL
-                 , configRedisMin    = defRedisMin
-                 , configRedisMax    = defRedisMax
-                 , configWipeMempool = defWipeMempool
-                 , configPeerTimeout = defPeerTimeout
-                 , configPeerMaxLife = defPeerMaxLife
-                 , configMaxPeers    = defMaxPeers
+    def = Config { configDir             = defDirectory
+                 , configHost            = defHost
+                 , configPort            = defPort
+                 , configNetwork         = defNetwork
+                 , configAsert           = defAsert
+                 , configDiscover        = defDiscover
+                 , configPeers           = defPeers
+                 , configVersion         = False
+                 , configDebug           = defDebug
+                 , configMaxPending      = defMaxPending
+                 , configWebLimits       = defWebLimits
+                 , configWebTimeouts     = defWebTimeouts
+                 , configRedis           = defRedis
+                 , configRedisURL        = defRedisURL
+                 , configRedisMin        = defRedisMin
+                 , configRedisMax        = defRedisMax
+                 , configWipeMempool     = defWipeMempool
+                 , configNoMempool       = defNoMempool
+                 , configPeerTimeout     = defPeerTimeout
+                 , configPeerMaxLife     = defPeerMaxLife
+                 , configMaxPeers        = defMaxPeers
+                 , configMaxDiff         = defMaxDiff
+                 , configCacheRefresh    = defCacheRefresh
+                 , configCacheRetryDelay = defCacheRetryDelay
+                 , configNumTxId         = defNumTxId
+                 , configStatsd          = defStatsd
+                 , configStatsdHost      = defStatsdHost
+                 , configStatsdPort      = defStatsdPort
+                 , configStatsdPrefix    = defStatsdPrefix
                  }
 
 defEnv :: MonadIO m => String -> a -> (String -> Maybe a) -> m a
@@ -93,6 +119,21 @@ defEnv e d p = do
     ms <- lookupEnv e
     return $ fromMaybe d $ p =<< ms
 
+defCacheRefresh :: Int
+defCacheRefresh = unsafePerformIO $
+    defEnv "CACHE_REFRESH" 750 readMaybe
+{-# NOINLINE defCacheRefresh #-}
+
+defCacheRetryDelay :: Int
+defCacheRetryDelay = unsafePerformIO $
+    defEnv "CACHE_RETRY_DELAY" 100000 readMaybe
+{-# NOINLINE defCacheRetryDelay #-}
+
+defMaxPending :: Int
+defMaxPending = unsafePerformIO $
+    defEnv "MAX_PENDING_TXS" 100 readMaybe
+{-# NOINLINE defMaxPending #-}
+
 defMaxPeers :: Int
 defMaxPeers = unsafePerformIO $
     defEnv "MAX_PEERS" 20 readMaybe
@@ -119,6 +160,11 @@ defNetwork = unsafePerformIO $
     defEnv "NET" bch (eitherToMaybe . networkReader)
 {-# NOINLINE defNetwork #-}
 
+defAsert :: Word32
+defAsert = unsafePerformIO $
+    defEnv "ASERT" 0 readMaybe
+{-# NOINLINE defAsert #-}
+
 defRedisMin :: Int
 defRedisMin = unsafePerformIO $
     defEnv "CACHE_MIN" 100 readMaybe
@@ -144,10 +190,10 @@ defDebug = unsafePerformIO $
     defEnv "DEBUG" False parseBool
 {-# NOINLINE defDebug #-}
 
-defReqLog :: Bool
-defReqLog = unsafePerformIO $
-    defEnv "REQ_LOG" False parseBool
-{-# NOINLINE defReqLog #-}
+defNumTxId :: Bool
+defNumTxId = unsafePerformIO $
+    defEnv "NUMTXID" False parseBool
+{-# NOINLINE defNumTxId #-}
 
 defWebLimits :: WebLimits
 defWebLimits = unsafePerformIO $ do
@@ -180,6 +226,11 @@ defWipeMempool = unsafePerformIO $
     defEnv "WIPE_MEMPOOL" False parseBool
 {-# NOINLINE defWipeMempool #-}
 
+defNoMempool :: Bool
+defNoMempool = unsafePerformIO $
+    defEnv "NO_MEMPOOL" False parseBool
+{-# NOINLINE defNoMempool #-}
+
 defRedisURL :: String
 defRedisURL = unsafePerformIO $
     defEnv "REDIS" "" pure
@@ -200,6 +251,41 @@ defPeerMaxLife = unsafePerformIO $
     defEnv "PEER_MAX_LIFE" (48 * 3600) readMaybe
 {-# NOINLINE defPeerMaxLife #-}
 
+defMaxDiff :: Int
+defMaxDiff = unsafePerformIO $
+    defEnv "MAX_DIFF" 2 readMaybe
+{-# NOINLINE defMaxDiff #-}
+
+defStatsd :: Bool
+defStatsd = unsafePerformIO $
+    defEnv "STATSD" False parseBool
+{-# NOINLINE defStatsd #-}
+
+defStatsdHost :: String
+defStatsdHost = unsafePerformIO $
+    defEnv "STATSD_HOST" "localhost" pure
+{-# NOINLINE defStatsdHost #-}
+
+defStatsdPort :: Int
+defStatsdPort = unsafePerformIO $
+    defEnv "STATSD_PORT" 8125 readMaybe
+{-# NOINLINE defStatsdPort #-}
+
+defStatsdPrefix :: String
+defStatsdPrefix =
+    unsafePerformIO $
+    runMaybeT go >>= \case
+        Nothing -> return "haskoin_store"
+        Just x -> return x
+  where
+    go = prefix <|> nomad
+    prefix = MaybeT $ lookupEnv "STATSD_PREFIX"
+    nomad = do
+        task <- MaybeT $ lookupEnv "NOMAD_TASK_NAME"
+        service <- MaybeT $ lookupEnv "NOMAD_ALLOC_INDEX"
+        return $ "app." <> task <> "." <> service
+{-# NOINLINE defStatsdPrefix #-}
+
 netNames :: String
 netNames = intercalate "|" (map getNetworkName allNets)
 
@@ -247,6 +333,12 @@ config = do
         <> help "Network to connect to"
         <> showDefault
         <> value (configNetwork def)
+    configAsert <-
+        option auto $
+        metavar "TIME"
+        <> long "asert"
+        <> help "ASERT (axon) activation time"
+        <> value (configAsert def)
     configDiscover <-
         flag (configDiscover def) True $
         long "discover"
@@ -263,6 +355,7 @@ config = do
         option auto $
         metavar "INT"
         <> long "max-peers"
+        <> help "Do not connect to more than this many peers"
         <> showDefault
         <> value (configMaxPeers def)
     configVersion <-
@@ -274,10 +367,6 @@ config = do
         flag (configDebug def) True $
         long "debug"
         <> help "Show debug messages"
-    configReqLog <-
-        flag (configReqLog def) True $
-        long "req-log"
-        <> help "HTTP request logging"
     maxLimitCount <-
         option auto $
         metavar "INT"
@@ -348,6 +437,13 @@ config = do
         <> help "Disconnect peers older than this"
         <> showDefault
         <> value (configPeerMaxLife def)
+    configMaxPending <-
+        option auto $
+        metavar "INT"
+        <> long "max-pending-txs"
+        <> help "Maximum pending txs to fail health check"
+        <> showDefault
+        <> value (configMaxPending def)
     configRedis <-
         flag (configRedis def) True $
         long "cache"
@@ -372,10 +468,64 @@ config = do
         <> help "Maximum number of keys in Redis xpub cache"
         <> showDefault
         <> value (configRedisMax def)
+    configCacheRefresh <-
+        option auto $
+        metavar "MILLISECONDS"
+        <> long "cache-refresh"
+        <> help "Refresh cache this frequently"
+        <> showDefault
+        <> value (configCacheRefresh def)
+    configCacheRetryDelay <-
+        option auto $
+        metavar "MICROSECONDS"
+        <> long "cache-retry-delay"
+        <> help "Delay to retry getting cache lock to index xpub"
+        <> showDefault
+        <> value (configCacheRetryDelay def)
+    configNoMempool <-
+        flag (configNoMempool def) True $
+        long "no-mempool"
+        <> help "Do not index new mempool transactions"
     configWipeMempool <-
         flag (configWipeMempool def) True $
         long "wipe-mempool"
         <> help "Wipe mempool at start"
+    configMaxDiff <-
+        option auto $
+        metavar "INT"
+        <> long "max-diff"
+        <> help "Maximum difference between headers and blocks"
+        <> showDefault
+        <> value (configMaxDiff def)
+    configNumTxId <-
+        flag (configNumTxId def) True $
+        long "numtxid"
+        <> help "Numeric tx_index field"
+    configStatsd <-
+        flag (configStatsd def) True $
+        long "statsd"
+        <> help "Enable statsd metrics"
+    configStatsdHost <-
+        strOption $
+        metavar "HOST"
+        <> long "statsd-host"
+        <> help "Host to send statsd metrics"
+        <> showDefault
+        <> value (configStatsdHost def)
+    configStatsdPort <-
+        option auto $
+        metavar "PORT"
+        <> long "statsd-port"
+        <> help "Port to send statsd metrics"
+        <> showDefault
+        <> value (configStatsdPort def)
+    configStatsdPrefix <-
+        strOption $
+        metavar "PREFIX"
+        <> long "statsd-prefix"
+        <> help "Prefix for statsd metrics"
+        <> showDefault
+        <> value (configStatsdPrefix def)
     pure
         Config
             { configWebLimits = WebLimits {..}
@@ -428,12 +578,13 @@ run :: Config -> IO ()
 run Config { configHost = host
            , configPort = port
            , configNetwork = net
+           , configAsert = asert
            , configDiscover = disc
            , configPeers = peers
            , configDir = db_dir
            , configDebug = deb
+           , configMaxPending = pend
            , configWebLimits = limits
-           , configReqLog = reqlog
            , configWebTimeouts = tos
            , configRedis = redis
            , configRedisURL = redisurl
@@ -443,8 +594,17 @@ run Config { configHost = host
            , configPeerTimeout = peertimeout
            , configPeerMaxLife = peerlife
            , configMaxPeers = maxpeers
+           , configMaxDiff = maxdiff
+           , configNoMempool = nomem
+           , configCacheRefresh = crefresh
+           , configCacheRetryDelay = cretrydelay
+           , configNumTxId = numtxid
+           , configStatsd = statsd
+           , configStatsdHost = statsdhost
+           , configStatsdPort = statsdport
+           , configStatsdPrefix = statsdpfx
            } =
-    runStderrLoggingT . filterLogger l $ do
+    runStderrLoggingT . filterLogger l . with_stats $ \stats -> do
         $(logInfoS) "Main" $
             "Creating working directory if not found: " <> cs wd
         createDirectoryIfMissing True wd
@@ -455,7 +615,7 @@ run Config { configHost = host
                           map (second (fromMaybe (getDefaultPort net))) peers
                     , storeConfDiscover = disc
                     , storeConfDB = wd </> "db"
-                    , storeConfNetwork = net
+                    , storeConfNetwork = net'
                     , storeConfCache =
                           if redis
                               then Just redisurl
@@ -465,9 +625,13 @@ run Config { configHost = host
                     , storeConfCacheMin = cachemin
                     , storeConfMaxKeys = redismax
                     , storeConfWipeMempool = wipemempool
+                    , storeConfNoMempool = nomem
                     , storeConfPeerTimeout = fromIntegral peertimeout
                     , storeConfPeerMaxLife = fromIntegral peerlife
                     , storeConfConnect = withConnection
+                    , storeConfCacheRefresh = crefresh
+                    , storeConfCacheRetryDelay = cretrydelay
+                    , storeConfStats = stats
                     }
         withStore scfg $ \st ->
             runWeb
@@ -476,12 +640,30 @@ run Config { configHost = host
                     , webPort = port
                     , webStore = st
                     , webMaxLimits = limits
-                    , webReqLog = reqlog
-                    , webWebTimeouts = tos
+                    , webTimeouts = tos
+                    , webMaxPending = pend
                     , webVersion = version
+                    , webMaxDiff = maxdiff
+                    , webNoMempool = nomem
+                    , webNumTxId = numtxid
+                    , webStats = stats
                     }
   where
+    with_stats go
+        | statsd = do
+            $(logInfoS) "Main" $
+                "Sending stats to " <> T.pack statsdhost <>
+                ":" <> cs (show statsdport) <>
+                " with prefix: " <> T.pack statsdpfx
+            withStats
+                (T.pack statsdhost)
+                statsdport
+                (T.pack statsdpfx)
+                (go . Just)
+        | otherwise = go Nothing
+    net' | asert == 0 = net
+         | otherwise = net { getAsertActivationTime = Just asert }
     l _ lvl
         | deb = True
         | otherwise = LevelInfo <= lvl
-    wd = db_dir </> getNetworkName net
+    wd = db_dir </> getNetworkName net'
