@@ -3,15 +3,14 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE LambdaCase        #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards   #-}
 {-# LANGUAGE TemplateHaskell   #-}
 {-# LANGUAGE TupleSections     #-}
 module Haskoin.Store.BlockStore
     ( -- * Block Store
       BlockStore
-    , BlockStoreMessage
-    , BlockStoreInbox
     , BlockStoreConfig(..)
-    , blockStore
+    , withBlockStore
     , blockStorePeerConnect
     , blockStorePeerConnectSTM
     , blockStorePeerDisconnect
@@ -26,6 +25,8 @@ module Haskoin.Store.BlockStore
     , blockStoreTxSTM
     , blockStoreTxHash
     , blockStoreTxHashSTM
+    , blockStorePendingTxs
+    , blockStorePendingTxsSTM
     ) where
 
 import           Control.Monad                 (forM, forM_, forever, mzero,
@@ -73,25 +74,29 @@ import           Haskoin.Node                  (Chain, OnlinePeer (..), Peer,
                                                 getPeers, killPeer, peerText,
                                                 sendMessage, setBusy, setFree)
 import           Haskoin.Store.Common
-import           Haskoin.Store.Data            (TxData (..), TxRef (..),
-                                                Unspent (..))
-import           Haskoin.Store.Database.Reader (DatabaseReader)
+import           Haskoin.Store.Data
+import           Haskoin.Store.Database.Reader
 import           Haskoin.Store.Database.Writer
 import           Haskoin.Store.Logic           (ImportException (Orphan),
                                                 deleteUnconfirmedTx,
                                                 getOldMempool, importBlock,
                                                 initBest, newMempoolTx,
                                                 revertBlock)
-import           NQE                           (Inbox, Listen, Mailbox,
-                                                Publisher, inboxToMailbox,
+import           Haskoin.Store.Stats
+import           NQE                           (Listen, Mailbox, Publisher,
+                                                inboxToMailbox, newInbox,
                                                 publish, query, receive, send,
                                                 sendSTM)
+import qualified System.Metrics                as Metrics
+import qualified System.Metrics.Gauge          as Metrics (Gauge)
+import qualified System.Metrics.Gauge          as Metrics.Gauge
 import           System.Random                 (randomRIO)
 import           UnliftIO                      (Exception, MonadIO,
                                                 MonadUnliftIO, STM, TVar, async,
-                                                atomically, liftIO, modifyTVar,
-                                                newTVarIO, readTVar, readTVarIO,
-                                                throwIO, withAsync, writeTVar)
+                                                atomically, liftIO, link,
+                                                modifyTVar, newTVarIO, readTVar,
+                                                readTVarIO, throwIO, withAsync,
+                                                writeTVar)
 import           UnliftIO.Concurrent           (threadDelay)
 
 data BlockStoreMessage
@@ -104,9 +109,6 @@ data BlockStoreMessage
     | TxRefAvailable !Peer ![TxHash]
     | BlockPing !(Listen ())
 
-type BlockStoreInbox = Inbox BlockStoreMessage
-type BlockStore = Mailbox BlockStoreMessage
-
 data BlockException
     = BlockNotInChain !BlockHash
     | Uninitialized
@@ -131,15 +133,80 @@ data PendingTx =
     deriving (Show, Eq, Ord)
 
 -- | Block store process state.
-data BlockRead =
-    BlockRead
-        { mySelf    :: !BlockStore
+data BlockStore =
+    BlockStore
+        { myMailbox :: !(Mailbox BlockStoreMessage)
         , myConfig  :: !BlockStoreConfig
         , myPeer    :: !(TVar (Maybe Syncing))
         , myTxs     :: !(TVar (HashMap TxHash PendingTx))
         , requested :: !(TVar (HashSet TxHash))
+        , myMetrics :: !(Maybe StoreMetrics)
         }
 
+data StoreMetrics = StoreMetrics
+    { storeHeight         :: !Metrics.Gauge
+    , headersHeight       :: !Metrics.Gauge
+    , storePendingTxs     :: !Metrics.Gauge
+    , storePeersConnected :: !Metrics.Gauge
+    , storeMempoolSize    :: !Metrics.Gauge
+    }
+
+newStoreMetrics :: MonadIO m => Metrics.Store -> m StoreMetrics
+newStoreMetrics s = liftIO $ do
+    storeHeight             <- g "height"
+    headersHeight           <- g "headers"
+    storePendingTxs         <- g "pending_txs"
+    storePeersConnected     <- g "peers_connected"
+    storeMempoolSize        <- g "mempool_size"
+    return StoreMetrics{..}
+  where
+    g x = Metrics.createGauge   ("store." <> x) s
+
+setStoreHeight :: MonadIO m => BlockT m ()
+setStoreHeight =
+    asks myMetrics >>= \case
+    Nothing -> return ()
+    Just m ->
+        getBestBlock >>= \case
+        Nothing -> setit m 0
+        Just bb -> getBlock bb >>= \case
+            Nothing -> setit m 0
+            Just b -> setit m (blockDataHeight b)
+  where
+    setit m i = liftIO $ storeHeight m `Metrics.Gauge.set` fromIntegral i
+
+setHeadersHeight :: MonadIO m => BlockT m ()
+setHeadersHeight =
+    asks myMetrics >>= \case
+    Nothing -> return ()
+    Just m -> do
+        h <- fmap nodeHeight $ chainGetBest =<< asks (blockConfChain . myConfig)
+        liftIO $ headersHeight m `Metrics.Gauge.set` fromIntegral h
+
+setPendingTxs :: MonadIO m => BlockT m ()
+setPendingTxs =
+    asks myMetrics >>= \case
+    Nothing -> return ()
+    Just m -> do
+        s <- asks myTxs >>= \t -> atomically (HashMap.size <$> readTVar t)
+        liftIO $ storePendingTxs m `Metrics.Gauge.set` fromIntegral s
+
+setPeersConnected :: MonadIO m => BlockT m ()
+setPeersConnected =
+    asks myMetrics >>= \case
+    Nothing -> return ()
+    Just m -> do
+        ps <- fmap length $ getPeers =<< asks (blockConfManager . myConfig)
+        liftIO $ storePeersConnected m `Metrics.Gauge.set` fromIntegral ps
+
+setMempoolSize :: MonadIO m => BlockT m ()
+setMempoolSize =
+    asks myMetrics >>= \case
+    Nothing -> return ()
+    Just m -> do
+        s <- length <$> getMempool
+        liftIO $ storeMempoolSize m `Metrics.Gauge.set` fromIntegral s
+
 -- | Configuration for a block store.
 data BlockStoreConfig =
     BlockStoreConfig
@@ -153,13 +220,16 @@ data BlockStoreConfig =
         -- ^ RocksDB database handle
         , blockConfNet         :: !Network
         -- ^ network constants
+        , blockConfNoMempool   :: !Bool
+        -- ^ do not index new mempool transactions
         , blockConfWipeMempool :: !Bool
         -- ^ wipe mempool at start
         , blockConfPeerTimeout :: !NominalDiffTime
         -- ^ disconnect syncing peer if inactive for this long
+        , blockConfStats       :: !(Maybe Metrics.Store)
         }
 
-type BlockT m = ReaderT BlockRead m
+type BlockT m = ReaderT BlockStore m
 
 runImport :: MonadLoggerIO m
           => WriterT (ExceptT ImportException m) a
@@ -206,30 +276,37 @@ instance MonadIO m => StoreReadExtra (BlockT m) where
         runRocksDB . getAddressTxs a
 
 -- | Run block store process.
-blockStore ::
+withBlockStore ::
        (MonadUnliftIO m, MonadLoggerIO m)
     => BlockStoreConfig
-    -> BlockStoreInbox
-    -> m ()
-blockStore cfg inbox = do
+    -> (BlockStore -> m a)
+    -> m a
+withBlockStore cfg action = do
     pb <- newTVarIO Nothing
     ts <- newTVarIO HashMap.empty
     rq <- newTVarIO HashSet.empty
-    runReaderT
-        (ini >> wipe >> run)
-        BlockRead { mySelf = inboxToMailbox inbox
-                  , myConfig = cfg
-                  , myPeer = pb
-                  , myTxs = ts
-                  , requested = rq
-                  }
+    inbox <- newInbox
+    metrics <- mapM newStoreMetrics (blockConfStats cfg)
+    let r = BlockStore { myMailbox = inboxToMailbox inbox
+                       , myConfig = cfg
+                       , myPeer = pb
+                       , myTxs = ts
+                       , requested = rq
+                       , myMetrics = metrics
+                       }
+    withAsync (runReaderT (go inbox) r) $ \a -> do
+        link a
+        action r
   where
+    go inbox = do
+        ini
+        wipe
+        run inbox
     del txs = do
         $(logInfoS) "BlockStore" $
             "Deleting " <> cs (show (length txs)) <> " transactions"
-        forM_ txs $ \tx ->
-            deleteUnconfirmedTx False (txRefHash tx)
-    wipeit txs = do
+        forM_ txs $ \(_, th) -> deleteUnconfirmedTx False th
+    wipe_it txs = do
         let (txs1, txs2) = splitAt 1000 txs
         unless (null txs1) $
             runImport (del txs1) >>= \case
@@ -237,10 +314,10 @@ blockStore cfg inbox = do
                     $(logErrorS) "BlockStore" $
                         "Could not wipe mempool: " <> cs (show e)
                     throwIO e
-                Right () -> wipeit txs2
+                Right () -> wipe_it txs2
     wipe
         | blockConfWipeMempool cfg =
-              getMempool >>= wipeit
+              getMempool >>= wipe_it
         | otherwise =
               return ()
     ini = runImport initBest >>= \case
@@ -249,7 +326,7 @@ blockStore cfg inbox = do
                       "Could not initialize: " <> cs (show e)
                   throwIO e
               Right () -> return ()
-    run = withAsync (pingMe (inboxToMailbox inbox))
+    run inbox = withAsync (pingMe (inboxToMailbox inbox))
           $ const
           $ forever
           $ receive inbox >>=
@@ -267,13 +344,20 @@ isInSync =
                 then clearSyncingState >> return True
                 else return False
 
-mempool :: MonadLoggerIO m => Peer -> m ()
-mempool p = do
-    $(logDebugS) "BlockStore" $
-        "Requesting mempool from peer: " <> peerText p
-    MMempool `sendMessage` p
+guardMempool :: Monad m => BlockT m () -> BlockT m ()
+guardMempool f = do
+    n <- asks (blockConfNoMempool . myConfig)
+    unless n f
 
-processBlock :: MonadLoggerIO m => Peer -> Block -> BlockT m ()
+mempool :: (MonadUnliftIO m, MonadLoggerIO m) => Peer -> BlockT m ()
+mempool p = guardMempool $ void $ async $ do
+    isInSync >>= \s -> when s $ do
+        $(logDebugS) "BlockStore" $
+            "Requesting mempool from peer: " <> peerText p
+        MMempool `sendMessage` p
+
+processBlock :: (MonadUnliftIO m, MonadLoggerIO m)
+             => Peer -> Block -> BlockT m ()
 processBlock peer block = void . runMaybeT $ do
     checkPeer peer >>= \case
         True -> return ()
@@ -297,7 +381,7 @@ processBlock peer block = void . runMaybeT $ do
         "Processing block: " <> blockText node Nothing
         <> " from peer: " <> peerText peer
     lift $ runImport (importBlock block node) >>= \case
-        Left e -> failure e
+        Left e   -> failure e
         Right () -> success node
   where
     header = blockHeader block
@@ -324,7 +408,7 @@ processBlock peer block = void . runMaybeT $ do
         listener <- asks (blockConfListener . myConfig)
         publish (StoreBestBlock blockhash) listener
 
-setSyncingBlocks :: (MonadReader BlockRead m, MonadIO m)
+setSyncingBlocks :: (MonadReader BlockStore m, MonadIO m)
                  => [BlockHash] -> m ()
 setSyncingBlocks hs =
     asks myPeer >>= \box ->
@@ -332,13 +416,13 @@ setSyncingBlocks hs =
         Nothing -> Nothing
         Just x  -> Just x { syncingBlocks = hs }
 
-getSyncingBlocks :: (MonadReader BlockRead m, MonadIO m) => m [BlockHash]
+getSyncingBlocks :: (MonadReader BlockStore m, MonadIO m) => m [BlockHash]
 getSyncingBlocks =
     asks myPeer >>= readTVarIO >>= \case
         Nothing -> return []
         Just x  -> return $ syncingBlocks x
 
-addSyncingBlocks :: (MonadReader BlockRead m, MonadIO m)
+addSyncingBlocks :: (MonadReader BlockStore m, MonadIO m)
                  => [BlockHash] -> m ()
 addSyncingBlocks hs =
     asks myPeer >>= \box ->
@@ -346,7 +430,7 @@ addSyncingBlocks hs =
         Nothing -> Nothing
         Just x  -> Just x { syncingBlocks = syncingBlocks x <> hs }
 
-removeSyncingBlock :: (MonadReader BlockRead m, MonadIO m)
+removeSyncingBlock :: (MonadReader BlockStore m, MonadIO m)
                    => BlockHash -> m ()
 removeSyncingBlock h = do
     box <- asks myPeer
@@ -354,13 +438,13 @@ removeSyncingBlock h = do
         Nothing -> Nothing
         Just x  -> Just x { syncingBlocks = delete h (syncingBlocks x) }
 
-checkPeer :: (MonadLoggerIO m, MonadReader BlockRead m) => Peer -> m Bool
+checkPeer :: (MonadLoggerIO m, MonadReader BlockStore m) => Peer -> m Bool
 checkPeer p =
     fmap syncingPeer <$> getSyncingState >>= \case
         Nothing -> return False
         Just p' -> return $ p == p'
 
-getBlockNode :: (MonadLoggerIO m, MonadReader BlockRead m)
+getBlockNode :: (MonadLoggerIO m, MonadReader BlockStore m)
              => BlockHash -> m (Maybe BlockNode)
 getBlockNode blockhash =
     chainGetBlock blockhash =<< asks (blockConfChain . myConfig)
@@ -382,7 +466,7 @@ processNoBlocks p hs = do
     killPeer (PeerMisbehaving "Did not find requested block(s)") p
 
 processTx :: MonadLoggerIO m => Peer -> Tx -> BlockT m ()
-processTx p tx = do
+processTx p tx = guardMempool $ do
     t <- liftIO getCurrentTime
     $(logDebugS) "BlockManager" $
         "Received tx " <> txHashToHex (txHash tx)
@@ -390,7 +474,7 @@ processTx p tx = do
     addPendingTx $ PendingTx t tx HashSet.empty
 
 pruneOrphans :: MonadIO m => BlockT m ()
-pruneOrphans = do
+pruneOrphans = guardMempool $ do
     ts <- asks myTxs
     now <- liftIO getCurrentTime
     atomically . modifyTVar ts . HashMap.filter $ \p ->
@@ -399,7 +483,12 @@ pruneOrphans = do
 addPendingTx :: MonadIO m => PendingTx -> BlockT m ()
 addPendingTx p = do
     ts <- asks myTxs
-    atomically $ modifyTVar ts $ HashMap.insert th p
+    rq <- asks requested
+    atomically $ do
+        modifyTVar ts $ HashMap.insert th p
+        modifyTVar rq $ HashSet.delete th
+        HashMap.size <$> readTVar ts
+    setPendingTxs
   where
     th = txHash (pendingTx p)
 
@@ -422,10 +511,13 @@ isPending th = do
               || th `HashSet.member` rs
 
 pendingTxs :: MonadIO m => Int -> BlockT m [PendingTx]
-pendingTxs i = asks myTxs >>= \box -> atomically $ do
-    pending <- readTVar box
-    let (selected, rest) = select pending
-    writeTVar box rest
+pendingTxs i = do
+    selected <- asks myTxs >>= \box -> atomically $ do
+        pending <- readTVar box
+        let (selected, rest) = select pending
+        writeTVar box rest
+        return (selected)
+    setPendingTxs
     return selected
   where
     select pend =
@@ -439,7 +531,7 @@ pendingTxs i = asks myTxs >>= \box -> atomically $ do
             txids = map (txHash . snd) sorted
          in mapMaybe (`HashMap.lookup` m) txids
 
-fulfillOrphans :: MonadIO m => BlockRead -> TxHash -> m ()
+fulfillOrphans :: MonadIO m => BlockStore -> TxHash -> m ()
 fulfillOrphans block_read th =
     atomically $ modifyTVar box (HashMap.map fulfill)
   where
@@ -449,7 +541,7 @@ fulfillOrphans block_read th =
 updateOrphans
     :: ( StoreReadBase m
        , MonadLoggerIO m
-       , MonadReader BlockRead m
+       , MonadReader BlockStore m
        )
     => m ()
 updateOrphans = do
@@ -479,7 +571,7 @@ updateOrphans = do
         return $ foldl fulfill p unspents
 
 newOrphanTx :: MonadLoggerIO m
-            => BlockRead
+            => BlockStore
             -> UTCTime
             -> Tx
             -> WriterT m ()
@@ -506,7 +598,7 @@ newOrphanTx block_read time tx = do
 
 importMempoolTx
     :: (MonadLoggerIO m, MonadError ImportException m)
-    => BlockRead
+    => BlockStore
     -> UTCTime
     -> Tx
     -> WriterT m Bool
@@ -534,7 +626,7 @@ importMempoolTx block_read time tx =
                 return False
 
 processMempool :: MonadLoggerIO m => BlockT m ()
-processMempool = do
+processMempool = guardMempool $ do
     txs <- pendingTxs 2000
     block_read <- ask
     unless (null txs) (import_txs block_read txs >>= success)
@@ -544,7 +636,7 @@ processMempool = do
             block_read
             (pendingTxTime p)
             (pendingTx p) >>= \case
-                True -> return $ Just (txHash (pendingTx p))
+                True  -> return $ Just (txHash (pendingTx p))
                 False -> return Nothing
     import_txs block_read txs =
         let r = mapM (run_import block_read) txs
@@ -565,7 +657,7 @@ processTxs ::
     => Peer
     -> [TxHash]
     -> BlockT m ()
-processTxs p hs = do
+processTxs p hs = guardMempool $ do
     s <- isInSync
     when s $ do
         $(logDebugS) "BlockStore" $
@@ -608,7 +700,7 @@ processTxs p hs = do
         msg `sendMessage` p
 
 touchPeer :: ( MonadIO m
-             , MonadReader BlockRead m
+             , MonadReader BlockStore m
              )
           => m ()
 touchPeer =
@@ -636,7 +728,7 @@ checkTime =
                 killPeer PeerTimeout p
 
 pruneMempool :: MonadLoggerIO m => BlockT m ()
-pruneMempool = do
+pruneMempool = guardMempool $ do
     s <- isInSync
     when s $ do
         now <- liftIO getCurrentTime
@@ -712,13 +804,11 @@ shouldSync :: MonadLoggerIO m => BlockT m (Maybe Peer)
 shouldSync =
     isInSync >>= \case
         True -> return Nothing
-        False -> cont
-  where
-    cont = getSyncingState >>= \case
-        Nothing -> return Nothing
-        Just Syncing { syncingPeer = p, syncingBlocks = bs }
-            | 100 > length bs -> return (Just p)
-            | otherwise -> return Nothing
+        False -> getSyncingState >>= \case
+            Nothing -> return Nothing
+            Just Syncing { syncingPeer = p, syncingBlocks = bs }
+                | 100 > length bs -> return (Just p)
+                | otherwise -> return Nothing
 
 syncMe :: MonadLoggerIO m => BlockT m ()
 syncMe = do
@@ -761,7 +851,7 @@ syncMe = do
         then return bh
         else findAncestor sh bh
 
-findAncestor :: (MonadLoggerIO m, MonadReader BlockRead m)
+findAncestor :: (MonadLoggerIO m, MonadReader BlockStore m)
              => BlockHeight -> BlockNode -> m BlockNode
 findAncestor height target = do
     ch <- asks (blockConfChain . myConfig)
@@ -775,13 +865,13 @@ findAncestor height target = do
                 <> cs (show (nodeHeight target))
             throwIO $ AncestorNotInChain height h
 
-finishPeer :: (MonadLoggerIO m, MonadReader BlockRead m)
+finishPeer :: (MonadLoggerIO m, MonadReader BlockStore m)
            => Peer -> m ()
 finishPeer p = do
     box <- asks myPeer
     readTVarIO box >>= \case
         Just Syncing { syncingPeer = p' } | p == p' -> reset_it box
-        _ -> return ()
+        _                                           -> return ()
   where
     reset_it box = do
         atomically $ writeTVar box Nothing
@@ -791,7 +881,7 @@ finishPeer p = do
 trySetPeer :: MonadLoggerIO m => Peer -> BlockT m Bool
 trySetPeer p =
     getSyncingState >>= \case
-        Just _ -> return False
+        Just _  -> return False
         Nothing -> set_it
   where
     set_it =
@@ -814,41 +904,41 @@ trySyncing =
     isInSync >>= \case
         True -> return ()
         False -> getSyncingState >>= \case
-            Just _ -> return ()
+            Just _  -> return ()
             Nothing -> online_peer
   where
     recurse [] = return ()
     recurse (p : ps) =
         trySetPeer p >>= \case
             False -> recurse ps
-            True -> syncMe
+            True  -> syncMe
     online_peer = do
         ops <- getPeers =<< asks (blockConfManager . myConfig)
         let ps = map onlinePeerMailbox ops
         recurse ps
 
-trySyncingPeer :: MonadLoggerIO m => Peer -> BlockT m ()
+trySyncingPeer :: (MonadUnliftIO m, MonadLoggerIO m) => Peer -> BlockT m ()
 trySyncingPeer p =
     isInSync >>= \case
-        True -> return ()
+        True -> mempool p
         False -> trySetPeer p >>= \case
             False -> return ()
-            True -> syncMe
+            True  -> syncMe
 
 getSyncingState
-    :: (MonadIO m, MonadReader BlockRead m) => m (Maybe Syncing)
+    :: (MonadIO m, MonadReader BlockStore m) => m (Maybe Syncing)
 getSyncingState =
     readTVarIO =<< asks myPeer
 
 clearSyncingState
-    :: (MonadLoggerIO m, MonadReader BlockRead m) => m ()
+    :: (MonadLoggerIO m, MonadReader BlockStore m) => m ()
 clearSyncingState =
     asks myPeer >>= readTVarIO >>= \case
-        Nothing -> return ()
+        Nothing                          -> return ()
         Just Syncing { syncingPeer = p } -> finishPeer p
 
-processBlockStoreMessage
-    :: MonadLoggerIO m => BlockStoreMessage -> BlockT m ()
+processBlockStoreMessage :: (MonadUnliftIO m, MonadLoggerIO m)
+                         => BlockStoreMessage -> BlockT m ()
 
 processBlockStoreMessage (BlockNewBest _) =
     trySyncing
@@ -877,9 +967,14 @@ processBlockStoreMessage (BlockPing r) = do
     pruneOrphans
     checkTime
     pruneMempool
+    setStoreHeight
+    setHeadersHeight
+    setPendingTxs
+    setPeersConnected
+    setMempoolSize
     atomically (r ())
 
-pingMe :: MonadLoggerIO m => BlockStore -> m ()
+pingMe :: MonadLoggerIO m => Mailbox BlockStoreMessage -> m ()
 pingMe mbox =
     forever $ do
         delay <- liftIO $
@@ -890,72 +985,84 @@ pingMe mbox =
 
 blockStorePeerConnect :: MonadIO m => Peer -> BlockStore -> m ()
 blockStorePeerConnect peer store =
-    BlockPeerConnect peer `send` store
+    BlockPeerConnect peer `send` myMailbox store
 
 blockStorePeerDisconnect
     :: MonadIO m => Peer -> BlockStore -> m ()
 blockStorePeerDisconnect peer store =
-    BlockPeerDisconnect peer `send` store
+    BlockPeerDisconnect peer `send` myMailbox store
 
 blockStoreHead
     :: MonadIO m => BlockNode -> BlockStore -> m ()
 blockStoreHead node store =
-    BlockNewBest node `send` store
+    BlockNewBest node `send` myMailbox store
 
 blockStoreBlock
     :: MonadIO m => Peer -> Block -> BlockStore -> m ()
 blockStoreBlock peer block store =
-    BlockReceived peer block `send` store
+    BlockReceived peer block `send` myMailbox store
 
 blockStoreNotFound
     :: MonadIO m => Peer -> [BlockHash] -> BlockStore -> m ()
 blockStoreNotFound peer blocks store =
-    BlockNotFound peer blocks `send` store
+    BlockNotFound peer blocks `send` myMailbox store
 
 blockStoreTx
     :: MonadIO m => Peer -> Tx -> BlockStore -> m ()
 blockStoreTx peer tx store =
-    TxRefReceived peer tx `send` store
+    TxRefReceived peer tx `send` myMailbox store
 
 blockStoreTxHash
     :: MonadIO m => Peer -> [TxHash] -> BlockStore -> m ()
 blockStoreTxHash peer txhashes store =
-    TxRefAvailable peer txhashes `send` store
+    TxRefAvailable peer txhashes `send` myMailbox store
 
 blockStorePeerConnectSTM
     :: Peer -> BlockStore -> STM ()
 blockStorePeerConnectSTM peer store =
-    BlockPeerConnect peer `sendSTM` store
+    BlockPeerConnect peer `sendSTM` myMailbox store
 
 blockStorePeerDisconnectSTM
     :: Peer -> BlockStore -> STM ()
 blockStorePeerDisconnectSTM peer store =
-    BlockPeerDisconnect peer `sendSTM` store
+    BlockPeerDisconnect peer `sendSTM` myMailbox store
 
 blockStoreHeadSTM
     :: BlockNode -> BlockStore -> STM ()
 blockStoreHeadSTM node store =
-    BlockNewBest node `sendSTM` store
+    BlockNewBest node `sendSTM` myMailbox store
 
 blockStoreBlockSTM
     :: Peer -> Block -> BlockStore -> STM ()
 blockStoreBlockSTM peer block store =
-    BlockReceived peer block `sendSTM` store
+    BlockReceived peer block `sendSTM` myMailbox store
 
 blockStoreNotFoundSTM
     :: Peer -> [BlockHash] -> BlockStore -> STM ()
 blockStoreNotFoundSTM peer blocks store =
-    BlockNotFound peer blocks `sendSTM` store
+    BlockNotFound peer blocks `sendSTM` myMailbox store
 
 blockStoreTxSTM
     :: Peer -> Tx -> BlockStore -> STM ()
 blockStoreTxSTM peer tx store =
-    TxRefReceived peer tx `sendSTM` store
+    TxRefReceived peer tx `sendSTM` myMailbox store
 
 blockStoreTxHashSTM
     :: Peer -> [TxHash] -> BlockStore -> STM ()
 blockStoreTxHashSTM peer txhashes store =
-    TxRefAvailable peer txhashes `sendSTM` store
+    TxRefAvailable peer txhashes `sendSTM` myMailbox store
+
+blockStorePendingTxs
+    :: MonadIO m => BlockStore -> m Int
+blockStorePendingTxs =
+    atomically . blockStorePendingTxsSTM
+
+blockStorePendingTxsSTM
+    :: BlockStore -> STM Int
+blockStorePendingTxsSTM BlockStore {..} = do
+    x <- HashMap.keysSet <$> readTVar myTxs
+    y <- readTVar requested
+    return $ HashSet.size $ x `HashSet.union` y
 
 blockText :: BlockNode -> Maybe Block -> Text
 blockText bn mblock = case mblock of
