@@ -22,49 +22,57 @@ module Haskoin.Store.Common
     , xPubBalsUnspents
     , getTransaction
     , blockAtOrBefore
+    , blockAtOrAfterMTP
     , deOffset
     , applyLimits
     , applyLimitsC
     , sortTxs
     , nub'
+    , microseconds
     ) where
 
-import           Conduit                   (ConduitT, dropC, mapC, takeC)
-import           Control.DeepSeq           (NFData)
-import           Control.Exception         (Exception)
-import           Control.Monad             (mzero)
-import           Control.Monad.Trans       (lift)
-import           Control.Monad.Trans.Maybe (MaybeT (..), runMaybeT)
-import           Data.ByteString           (ByteString)
-import           Data.Default              (Default (..))
-import           Data.Function             (on)
-import           Data.Hashable             (Hashable)
-import qualified Data.HashSet              as H
-import           Data.IntMap.Strict        (IntMap)
-import qualified Data.IntMap.Strict        as I
-import           Data.List                 (sortBy)
-import           Data.Maybe                (catMaybes, listToMaybe)
-import           Data.Serialize            (Serialize (..))
-import           Data.Word                 (Word32, Word64)
-import           GHC.Generics              (Generic)
-import           Haskoin                   (Address, BlockHash,
-                                            BlockHeader (..), BlockHeight,
-                                            KeyIndex, Network (..),
-                                            OutPoint (..), RejectCode (..),
-                                            Tx (..), TxHash (..), TxIn (..),
-                                            XPubKey (..), deriveAddr,
-                                            deriveCompatWitnessAddr,
-                                            deriveWitnessAddr, pubSubKey,
-                                            txHash)
-import           Haskoin.Node              (Peer)
-import           Haskoin.Store.Data        (Balance (..), BlockData (..),
-                                            DeriveType (..), Spender,
-                                            Transaction, TxData (..),
-                                            TxRef (..), UnixTime, Unspent (..),
-                                            XPubBal (..), XPubSpec (..),
-                                            XPubSummary (..), XPubUnspent (..),
-                                            nullBalance, toTransaction,
-                                            zeroBalance)
+import           Conduit                    (ConduitT, dropC, mapC, takeC)
+import           Control.DeepSeq            (NFData)
+import           Control.Exception          (Exception)
+import           Control.Monad.Trans        (lift)
+import           Control.Monad.Trans.Maybe  (MaybeT (..), runMaybeT)
+import           Control.Monad.Trans.Reader (runReaderT)
+import           Data.ByteString            (ByteString)
+import           Data.Default               (Default (..))
+import           Data.Function              (on)
+import           Data.Hashable              (Hashable)
+import qualified Data.HashSet               as H
+import           Data.IntMap.Strict         (IntMap)
+import qualified Data.IntMap.Strict         as I
+import           Data.List                  (sortBy)
+import           Data.Maybe                 (catMaybes)
+import           Data.Serialize             (Serialize (..))
+import           Data.Time.Clock.System     (getSystemTime, systemNanoseconds,
+                                             systemSeconds)
+import           Data.Word                  (Word32, Word64)
+import           GHC.Generics               (Generic)
+import           Haskoin                    (Address, BlockHash,
+                                             BlockHeader (..), BlockHeight,
+                                             BlockNode (..), KeyIndex,
+                                             Network (..), OutPoint (..),
+                                             RejectCode (..), Tx (..),
+                                             TxHash (..), TxIn (..),
+                                             XPubKey (..), deriveAddr,
+                                             deriveCompatWitnessAddr,
+                                             deriveWitnessAddr,
+                                             firstGreaterOrEqual, headerHash,
+                                             lastSmallerOrEqual, mtp, pubSubKey,
+                                             txHash)
+import           Haskoin.Node               (Chain, Peer)
+import           Haskoin.Store.Data         (Balance (..), BlockData (..),
+                                             DeriveType (..), Spender,
+                                             Transaction, TxData (..),
+                                             TxRef (..), UnixTime, Unspent (..),
+                                             XPubBal (..), XPubSpec (..),
+                                             XPubSummary (..), XPubUnspent (..),
+                                             nullBalance, toTransaction,
+                                             zeroBalance)
+import           UnliftIO                   (MonadIO, liftIO)
 
 type DeriveAddr = XPubKey -> KeyIndex -> Address
 
@@ -101,7 +109,7 @@ class Monad m => StoreReadBase m where
     getSpender :: OutPoint -> m (Maybe Spender)
     getBalance :: Address -> m (Maybe Balance)
     getUnspent :: OutPoint -> m (Maybe Unspent)
-    getMempool :: m [TxRef]
+    getMempool :: m [(UnixTime, TxHash)]
 
 class StoreReadBase m => StoreReadExtra m where
     getBalances :: [Address] -> m [Balance]
@@ -194,7 +202,8 @@ class StoreWrite m where
     deleteAddrTx :: Address -> TxRef -> m ()
     insertAddrUnspent :: Address -> Unspent -> m ()
     deleteAddrUnspent :: Address -> Unspent -> m ()
-    setMempool :: [TxRef] -> m ()
+    addToMempool :: TxHash -> UnixTime -> m ()
+    deleteFromMempool :: TxHash -> m ()
     setBalance :: Balance -> m ()
     insertUnspent :: Unspent -> m ()
     deleteUnspent :: OutPoint -> m ()
@@ -203,17 +212,10 @@ getSpenders :: StoreReadBase m => TxHash -> m (IntMap Spender)
 getSpenders th =
     getActiveTxData th >>= \case
         Nothing -> return I.empty
-        Just td ->
-            I.fromList . catMaybes . zipWith f [0..]
-            <$> mapM getSpender (ops td)
-
+        Just td -> I.fromList . catMaybes <$>
+                   mapM get_spender [0 .. length (txOut (txData td)) - 1]
   where
-    f i m = (i, ) <$> m
-    ops td =
-        let tx = txData td
-            outs = txOut tx
-            op i _ = OutPoint th i
-         in zipWith op [0..] outs
+    get_spender i = fmap (i,) <$> getSpender (OutPoint th (fromIntegral i))
 
 getActiveBlock :: StoreReadExtra m => BlockHash -> m (Maybe BlockData)
 getActiveBlock bh = getBlock bh >>= \case
@@ -275,25 +277,31 @@ getTransaction h = runMaybeT $ do
     sm <- lift $ getSpenders h
     return $ toTransaction d sm
 
-blockAtOrBefore :: (Monad m, StoreReadExtra m)
-                => UnixTime
+blockAtOrBefore :: (MonadIO m, StoreReadExtra m)
+                => Chain
+                -> UnixTime
                 -> m (Maybe BlockData)
-blockAtOrBefore q = runMaybeT $ do
-    a <- g 0
-    b <- MaybeT getBestBlock >>= MaybeT . getBlock
-    f a b
+blockAtOrBefore ch q = runMaybeT $ do
+    net <- lift getNetwork
+    x <- MaybeT $ liftIO $ runReaderT (lastSmallerOrEqual net f) ch
+    MaybeT $ getBlock (headerHash (nodeHeader x))
   where
-    f a b
-        | t b <= q = return b
-        | t a > q = mzero
-        | h b - h a == 1 = return a
-        | otherwise = do
-              let x = h a + (h b - h a) `div` 2
-              m <- g x
-              if t m > q then f a m else f m b
-    g x = MaybeT (listToMaybe <$> getBlocksAtHeight x) >>= MaybeT . getBlock
-    h = blockDataHeight
-    t = fromIntegral . blockTimestamp . blockDataHeader
+    f x =
+        let t = fromIntegral (blockTimestamp (nodeHeader x))
+         in return $ t `compare` q
+
+blockAtOrAfterMTP :: (MonadIO m, StoreReadExtra m)
+                  => Chain
+                  -> UnixTime
+                  -> m (Maybe BlockData)
+blockAtOrAfterMTP ch q = runMaybeT $ do
+    net <- lift getNetwork
+    x <- MaybeT $ liftIO $ runReaderT (firstGreaterOrEqual net f) ch
+    MaybeT $ getBlock (headerHash (nodeHeader x))
+  where
+    f x = do
+        t <- fromIntegral <$> mtp x
+        return $ t `compare` q
 
 
 -- | Events that the store can generate.
@@ -305,7 +313,6 @@ data StoreEvent
     | StorePeerPong !Peer !Word64
     | StoreTxAvailable !Peer ![TxHash]
     | StoreTxReject !Peer !TxHash !RejectCode !ByteString
-      -- ^ block no longer head of main chain
 
 data PubExcept = PubNoPeers
     | PubReject RejectCode
@@ -369,3 +376,9 @@ sortTxs txs = go [] thset $ zip [0 ..] txs
 
 nub' :: (Eq a, Hashable a) => [a] -> [a]
 nub' = H.toList . H.fromList
+
+microseconds :: MonadIO m => m Integer
+microseconds =
+    let f t = toInteger (systemSeconds t) * 1000000
+            + toInteger (systemNanoseconds t) `div` 1000
+    in liftIO $ f <$> getSystemTime
