@@ -10,7 +10,6 @@ module Haskoin.Store.Database.Types
     , BalKey(..)
     , HeightKey(..)
     , MemKey(..)
-    , OldMemKey(..)
     , SpenderKey(..)
     , TxKey(..)
     , UnspentKey(..)
@@ -27,24 +26,25 @@ module Haskoin.Store.Database.Types
 
 import           Control.DeepSeq        (NFData)
 import           Control.Monad          (guard)
+import           Data.Bits              (Bits, shiftL, shiftR, (.|.))
 import           Data.ByteString        (ByteString)
 import qualified Data.ByteString        as BS
 import           Data.ByteString.Short  (ShortByteString)
 import qualified Data.ByteString.Short  as BSS
 import           Data.Default           (Default (..))
+import           Data.Either            (fromRight)
 import           Data.Hashable          (Hashable)
-import           Data.Serialize         (Serialize (..), getBytes, getWord8,
-                                         putWord8)
-import           Data.Word              (Word32, Word64)
+import           Data.Serialize         (Serialize (..), decode, encode,
+                                         getBytes, getWord8, putWord8, runPut)
+import           Data.Word              (Word16, Word32, Word64, Word8)
 import           Database.RocksDB.Query (Key, KeyValue)
 import           GHC.Generics           (Generic)
 import           Haskoin                (Address, BlockHash, BlockHeight,
                                          OutPoint (..), TxHash, eitherToMaybe,
                                          scriptToAddressBS)
 import           Haskoin.Store.Data     (Balance (..), BlockData, BlockRef,
-                                         TxRef (..), Spender, TxData,
-                                         UnixTime, Unspent (..), getUnixTime,
-                                         putUnixTime)
+                                         Spender, TxData, TxRef (..), UnixTime,
+                                         Unspent (..))
 
 -- | Database key for an address transaction.
 data AddrTxKey
@@ -136,9 +136,8 @@ data OutVal = OutVal
 instance KeyValue AddrOutKey OutVal
 
 -- | Transaction database key.
-newtype TxKey = TxKey
-    { txKey :: TxHash
-    } deriving (Show, Read, Eq, Ord, Generic, Hashable)
+data TxKey = TxKey { txKey :: TxHash }
+    deriving (Show, Read, Eq, Ord, Generic, Hashable)
 
 instance Serialize TxKey where
     -- 0x02 · TxHash
@@ -241,6 +240,7 @@ instance Serialize BlockKey where
         guard . (== 0x01) =<< getWord8
         BlockKey <$> get
 
+instance Key BlockKey
 instance KeyValue BlockKey BlockData
 
 -- | Block height database key.
@@ -312,38 +312,6 @@ instance Serialize VersionKey where
 instance Key VersionKey
 instance KeyValue VersionKey Word32
 
-
--- | Old mempool transaction database key.
-data OldMemKey
-    = OldMemKey
-          { memTime :: !UnixTime
-          , memKey  :: !TxHash
-          }
-    | OldMemKeyT
-          { memTime :: !UnixTime
-          }
-    | OldMemKeyS
-    deriving (Show, Read, Eq, Ord, Generic, Hashable)
-
-instance Serialize OldMemKey where
-    -- 0x07 · UnixTime · TxHash
-    put (OldMemKey t h) = do
-        putWord8 0x07
-        putUnixTime t
-        put h
-    -- 0x07 · UnixTime
-    put (OldMemKeyT t) = do
-        putWord8 0x07
-        putUnixTime t
-    -- 0x07
-    put OldMemKeyS = putWord8 0x07
-    get = do
-        guard . (== 0x07) =<< getWord8
-        OldMemKey <$> getUnixTime <*> get
-
-instance Key OldMemKey
-instance KeyValue OldMemKey ()
-
 data BalVal = BalVal
     { balValAmount        :: !Word64
     , balValZero          :: !Word64
