@@ -16,57 +16,42 @@ module Haskoin.Store.Logic
     , deleteUnconfirmedTx
     ) where
 
-import           Control.Monad                 (forM_, guard, unless, void,
-                                                when, zipWithM_, (<=<))
-import           Control.Monad.Except          (ExceptT (..), MonadError,
-                                                runExceptT, throwError)
-import           Control.Monad.Logger          (LoggingT (..),
-                                                MonadLoggerIO (..), logDebugS,
-                                                logErrorS)
-import           Control.Monad.Reader          (ReaderT (ReaderT), runReaderT)
-import           Control.Monad.Trans           (lift)
-import qualified Data.ByteString               as B
-import qualified Data.ByteString.Short         as B.Short
-import           Data.Either                   (rights)
-import qualified Data.HashSet                  as S
-import qualified Data.IntMap.Strict            as I
-import           Data.List                     (nub, sortOn)
-import           Data.Maybe                    (catMaybes, fromMaybe, isJust,
-                                                isNothing, mapMaybe)
-import           Data.Ord                      (Down (Down))
-import           Data.Serialize                (encode)
-import           Data.String                   (IsString (..))
-import           Data.String.Conversions       (ConvertibleStrings, cs)
-import           Data.Text                     (Text)
-import           Data.Word                     (Word32, Word64)
-import           Haskoin                       (Address, Block (..), BlockHash,
-                                                BlockHeader (..),
-                                                BlockNode (..), Network (..),
-                                                OutPoint (..), Tx (..), TxHash,
-                                                TxIn (..), TxOut (..),
-                                                blockHashToHex, computeSubsidy,
-                                                eitherToMaybe, genesisBlock,
-                                                genesisNode, headerHash,
-                                                isGenesis, nullOutPoint,
-                                                scriptToAddressBS, txHash,
-                                                txHashToHex)
+import           Control.Monad         (forM, forM_, guard, unless, void, when,
+                                        zipWithM_)
+import           Control.Monad.Except  (MonadError, throwError)
+import           Control.Monad.Logger  (MonadLoggerIO (..), logDebugS,
+                                        logErrorS)
+import qualified Data.ByteString       as B
+import qualified Data.ByteString.Short as B.Short
+import           Data.Either           (rights)
+import qualified Data.IntMap.Strict    as I
+import           Data.List             (nub)
+import           Data.Maybe            (catMaybes, fromMaybe, isJust, isNothing)
+import           Data.Serialize        (encode)
+import           Data.Word             (Word32, Word64)
+import           Haskoin               (Address, Block (..), BlockHash,
+                                        BlockHeader (..), BlockNode (..),
+                                        Network (..), OutPoint (..), Tx (..),
+                                        TxHash, TxIn (..), TxOut (..),
+                                        blockHashToHex, computeSubsidy,
+                                        eitherToMaybe, genesisBlock,
+                                        genesisNode, headerHash, isGenesis,
+                                        nullOutPoint, scriptToAddressBS, txHash,
+                                        txHashToHex)
 import           Haskoin.Store.Common
-import           Haskoin.Store.Data            (Balance (..), BlockData (..),
-                                                BlockRef (..), Prev (..),
-                                                Spender (..), TxData (..),
-                                                TxRef (..), UnixTime,
-                                                Unspent (..), confirmed)
-import           Haskoin.Store.Database.Writer (MemoryTx, WriterT, runTx)
-import           UnliftIO                      (Exception, MonadIO, async,
-                                                liftIO, wait)
+import           Haskoin.Store.Data    (Balance (..), BlockData (..),
+                                        BlockRef (..), Prev (..), Spender (..),
+                                        TxData (..), TxRef (..), UnixTime,
+                                        Unspent (..), confirmed)
+import           UnliftIO              (Exception)
 
 type MonadImport m =
     ( MonadError ImportException m
     , MonadLoggerIO m
+    , StoreReadBase m
+    , StoreWrite m
     )
 
-type MonadMemory = ExceptT ImportException MemoryTx
-
 data ImportException
     = PrevBlockNotBest
     | Orphan
@@ -96,13 +81,7 @@ instance Show ImportException where
     show TxSpent             = "Transaction is spent"
     show OrphanLoop          = "Orphan loop"
 
-runMonadMemory :: MonadImport m => MonadMemory a -> WriterT m a
-runMonadMemory f =
-    runTx (runExceptT f) >>= \case
-        Left e -> throwError e
-        Right x -> return x
-
-initBest :: MonadImport m => WriterT m ()
+initBest :: MonadImport m => m ()
 initBest = do
     $(logDebugS) "BlockStore" "Initializing best block"
     net <- getNetwork
@@ -113,36 +92,21 @@ initBest = do
 
 getOldMempool :: StoreReadBase m => UnixTime -> m [TxHash]
 getOldMempool now =
-    map txRefHash . filter f <$> getMempool
-  where
-    f = (< now - 3600 * 72) . memRefTime . txRefBlock
+    map snd . filter ((< now - 3600 * 72) . fst) <$> getMempool
 
-newMempoolTx :: MonadImport m => Tx -> UnixTime -> WriterT m Bool
+newMempoolTx :: MonadImport m => Tx -> UnixTime -> m Bool
 newMempoolTx tx w =
     getActiveTxData (txHash tx) >>= \case
-        Just _ -> return False
+        Just _ ->
+            return False
         Nothing -> do
-            freeOutputs True True [tx]
-            preLoadMemory [tx]
+            freeOutputs True True tx
             rbf <- isRBF (MemRef w) tx
             checkNewTx tx
-            runMonadMemory $ importTx (MemRef w) w rbf tx
+            importTx (MemRef w) w rbf tx
             return True
 
-preLoadMemory :: MonadLoggerIO m => [Tx] -> WriterT m ()
-preLoadMemory txs = do
-    $(logDebugS) "BlockStore" "Pre-loading memory"
-    ReaderT $ \r -> do
-        l <- askLoggerIO
-        liftIO (runLoggingT (runReaderT go r) l)
-  where
-    go = do
-        prev_asyncs <- mapM (async . mapM_ loadPrevOut . prevOuts) txs
-        out_asyncs <- mapM (async . loadOutputBalances) txs
-        mapM_ wait $ prev_asyncs <> out_asyncs
-        runTx . setMempool =<< getMempool
-
-bestBlockData :: MonadImport m => WriterT m BlockData
+bestBlockData :: MonadImport m => m BlockData
 bestBlockData = do
     h <- getBestBlock >>= \case
         Nothing -> do
@@ -155,7 +119,7 @@ bestBlockData = do
             throwError BestBlockNotFound
         Just b -> return b
 
-revertBlock :: MonadImport m => BlockHash -> WriterT m ()
+revertBlock :: MonadImport m => BlockHash -> m ()
 revertBlock bh = do
     bd <- bestBlockData >>= \b ->
         if headerHash (blockDataHeader b) == bh
@@ -165,14 +129,12 @@ revertBlock bh = do
                 "Cannot revert non-head block: " <> blockHashToHex bh
             throwError BlockNotBest
     tds <- mapM getImportTxData (blockDataTxs bd)
-    preLoadMemory $ map txData tds
-    runTx $ do
-        setBest (prevBlock (blockDataHeader bd))
-        insertBlock bd {blockDataMainChain = False}
-        forM_ (tail tds) unConfirmTx
+    setBest (prevBlock (blockDataHeader bd))
+    insertBlock bd {blockDataMainChain = False}
+    forM_ (tail tds) unConfirmTx
     deleteConfirmedTx (txHash (txData (head tds)))
 
-checkNewBlock :: MonadImport m => Block -> BlockNode -> WriterT m ()
+checkNewBlock :: MonadImport m => Block -> BlockNode -> m ()
 checkNewBlock b n =
     getBestBlock >>= \case
         Nothing
@@ -190,12 +152,12 @@ checkNewBlock b n =
                     <> blockHashToHex (headerHash (blockHeader b))
                 throwError PrevBlockNotBest
 
-importOrConfirm :: MonadImport m => BlockNode -> [Tx] -> WriterT m ()
-importOrConfirm bn txs = do
-    freeOutputs True False txs
-    preLoadMemory txs
-    mapM_ (uncurry action) (sortTxs txs)
+importOrConfirm :: MonadImport m => BlockNode -> [Tx] -> m ()
+importOrConfirm bn txns = do
+    mapM_ (freeOutputs True False . snd) (reverse txs)
+    mapM_ (uncurry action) txs
   where
+    txs = sortTxs txns
     br i = BlockRef {blockRefHeight = nodeHeight bn, blockRefPos = i}
     bn_time = fromIntegral . blockTimestamp $ nodeHeader bn
     action i tx =
@@ -208,7 +170,7 @@ importOrConfirm bn txs = do
                 $(logDebugS) "BlockStore" $
                     "Confirming tx: "
                     <> txHashToHex (txHash tx)
-                runTx $ confTx t (Just (br i))
+                confirmTx t (br i)
                 return Nothing
             Nothing -> do
                 $(logErrorS) "BlockStore" $
@@ -218,10 +180,10 @@ importOrConfirm bn txs = do
     import_it i tx = do
         $(logDebugS) "BlockStore" $
             "Importing tx: " <> txHashToHex (txHash tx)
-        runMonadMemory $ importTx (br i) bn_time False tx
+        importTx (br i) bn_time False tx
         return Nothing
 
-importBlock :: MonadImport m => Block -> BlockNode -> WriterT m ()
+importBlock :: MonadImport m => Block -> BlockNode -> m ()
 importBlock b n = do
     $(logDebugS) "BlockStore" $
         "Checking new block: "
@@ -234,24 +196,23 @@ importBlock b n = do
     $(logDebugS) "BlockStore" $
         "Inserting block entries for: "
         <> blockHashToHex (headerHash (nodeHeader n))
-    runTx $ do
-        insertBlock
-            BlockData
-                { blockDataHeight = nodeHeight n
-                , blockDataMainChain = True
-                , blockDataWork = nodeWork n
-                , blockDataHeader = nodeHeader n
-                , blockDataSize = fromIntegral (B.length (encode b))
-                , blockDataTxs = map txHash (blockTxns b)
-                , blockDataWeight = if getSegWit net then w else 0
-                , blockDataSubsidy = subsidy
-                , blockDataFees = cb_out_val - subsidy
-                , blockDataOutputs = ts_out_val
-                }
-        setBlocksAtHeight
-            (nub (headerHash (nodeHeader n) : bs))
-            (nodeHeight n)
-        setBest (headerHash (nodeHeader n))
+    insertBlock
+        BlockData
+            { blockDataHeight = nodeHeight n
+            , blockDataMainChain = True
+            , blockDataWork = nodeWork n
+            , blockDataHeader = nodeHeader n
+            , blockDataSize = fromIntegral (B.length (encode b))
+            , blockDataTxs = map txHash (blockTxns b)
+            , blockDataWeight = if getSegWit net then w else 0
+            , blockDataSubsidy = subsidy
+            , blockDataFees = cb_out_val - subsidy
+            , blockDataOutputs = ts_out_val
+            }
+    setBlocksAtHeight
+        (nub (headerHash (nodeHeader n) : bs))
+        (nodeHeight n)
+    setBest (headerHash (nodeHeader n))
     importOrConfirm n (blockTxns b)
     $(logDebugS) "BlockStore" $
         "Finished importing transactions for: "
@@ -268,14 +229,18 @@ importBlock b n = do
             s = B.length (encode b')
          in fromIntegral $ s * 3 + x
 
-checkNewTx :: MonadImport m => Tx -> WriterT m ()
+checkNewTx :: MonadImport m => Tx -> m ()
 checkNewTx tx = do
-    us <- runTx $ getUnspentOutputs tx
     when (unique_inputs < length (txIn tx)) $ do
         $(logErrorS) "BlockStore" $
             "Transaction spends same output twice: "
             <> txHashToHex (txHash tx)
         throwError DuplicatePrevOutput
+    us <- getUnspentOutputs tx
+    when (any isNothing us) $ do
+        $(logErrorS) "BlockStore" $
+            "Orphan: " <> txHashToHex (txHash tx)
+        throwError Orphan
     when (isCoinbase tx) $ do
         $(logErrorS) "BlockStore" $
             "Coinbase cannot be imported into mempool: "
@@ -290,50 +255,15 @@ checkNewTx tx = do
             "Insufficient funds for tx: " <> txHashToHex (txHash tx)
         throwError InsufficientFunds
   where
-    unspents = sum . map unspentAmount
+    unspents = sum . map unspentAmount . catMaybes
     outputs = sum (map outValue (txOut tx))
     unique_inputs = length (nub' (map prevOutput (txIn tx)))
 
-getUnspentOutputs :: StoreReadBase m => Tx -> m [Unspent]
-getUnspentOutputs tx = catMaybes <$> mapM getUnspent (prevOuts tx)
+getUnspentOutputs :: StoreReadBase m => Tx -> m [Maybe Unspent]
+getUnspentOutputs tx = mapM getUnspent (prevOuts tx)
 
-loadUnspentOutput :: MonadLoggerIO m => Unspent -> WriterT m ()
-loadUnspentOutput u = do
-    mbal <- mapM getDefaultBalance (unspentAddress u)
-    runTx $ do
-        insertUnspent u
-        mapM_ setBalance mbal
-
-loadTxData :: MonadLoggerIO m => TxData -> WriterT m ()
-loadTxData td = do
-    runTx (insertTx td)
-    bals <- mapM getDefaultBalance addrs
-    ss <- getSpenders (txHash (txData td))
-    runTx $ do
-        mapM_ setBalance bals
-        forM_ (I.toList ss) $ \(i, s) -> do
-            let op = OutPoint (txHash (txData td)) (fromIntegral i)
-            insertSpender op s
-  where
-    addrs =
-        let f = eitherToMaybe . scriptToAddressBS . scriptOutput
-        in mapMaybe f (txOut (txData td))
-
-loadPrevOut :: MonadLoggerIO m => OutPoint -> WriterT m ()
-loadPrevOut op = getUnspent op >>= \case
-    Just u -> loadUnspentOutput u
-    Nothing -> getActiveTxData (outPointHash op) >>= \case
-        Just td -> loadTxData td
-        Nothing -> return ()
-
-loadOutputBalances :: MonadIO m => Tx -> WriterT m ()
-loadOutputBalances tx = do
-    let f = eitherToMaybe . scriptToAddressBS . scriptOutput
-    let addrs = mapMaybe f (txOut tx)
-    forM_ addrs $ runTx . setBalance <=< getDefaultBalance
-
-prepareTxData :: Bool -> BlockRef -> Word64 -> [Unspent] -> Tx -> TxData
-prepareTxData rbf br tt us tx =
+prepareTxData :: Bool -> BlockRef -> Word64 -> Tx -> [Unspent] -> TxData
+prepareTxData rbf br tt tx us =
     TxData { txDataBlock = br
            , txData = tx
            , txDataPrevs = ps
@@ -343,23 +273,34 @@ prepareTxData rbf br tt us tx =
            }
   where
     mkprv u = Prev (B.Short.fromShort (unspentScript u)) (unspentAmount u)
-    ps = I.fromList $ zip [0 ..] $ if isCoinbase tx then [] else map mkprv us
+    ps = I.fromList $ zip [0 ..] $ map mkprv us
 
 importTx
-    :: BlockRef
+    :: MonadImport m
+    => BlockRef
     -> Word64 -- ^ unix time
     -> Bool -- ^ RBF
     -> Tx
-    -> MonadMemory ()
+    -> m ()
 importTx br tt rbf tx = do
-    us <- lift $ getUnspentOutputs tx
-    let td = prepareTxData rbf br tt us tx
-    lift $ commitAddTx us td
+    mus <- getUnspentOutputs tx
+    us <- forM mus $ \case
+        Nothing -> do
+            $(logErrorS) "BlockStore" $
+                "Attempted to import a tx missing UTXO: "
+                <> txHashToHex (txHash tx)
+            throwError Orphan
+        Just u -> return u
+    let td = prepareTxData rbf br tt tx us
+    commitAddTx td
 
-unConfirmTx :: TxData -> MemoryTx ()
+unConfirmTx :: MonadImport m => TxData -> m ()
 unConfirmTx t = confTx t Nothing
 
-replaceAddressTx :: TxData -> BlockRef -> MemoryTx ()
+confirmTx :: MonadImport m => TxData -> BlockRef -> m ()
+confirmTx t br = confTx t (Just br)
+
+replaceAddressTx :: MonadImport m => TxData -> BlockRef -> m ()
 replaceAddressTx t new = forM_ (txDataAddresses t) $ \a -> do
     deleteAddrTx
         a
@@ -370,11 +311,16 @@ replaceAddressTx t new = forM_ (txDataAddresses t) $ \a -> do
         TxRef { txRefBlock = new
               , txRefHash = txHash (txData t) }
 
-adjustAddressOutput :: OutPoint -> TxOut -> BlockRef -> BlockRef -> MemoryTx ()
+adjustAddressOutput :: MonadImport m
+                    => OutPoint -> TxOut -> BlockRef -> BlockRef -> m ()
 adjustAddressOutput op o old new = do
     let pk = scriptOutput o
-    u <- getUnspent op
-    when (isJust u) $ replace_unspent pk
+    getUnspent op >>= \case
+        Nothing -> return ()
+        Just u -> do
+            unless (unspentBlock u == old) $
+                error $ "Existing unspent block bad for output: " <> show op
+            replace_unspent pk
   where
     replace_unspent pk = do
         let ma = eitherToMaybe (scriptToAddressBS pk)
@@ -410,72 +356,57 @@ adjustAddressOutput op o old new = do
         decreaseBalance (confirmed old) a (outValue o)
         increaseBalance (confirmed new) a (outValue o)
 
-confTx :: TxData -> Maybe BlockRef -> MemoryTx ()
+confTx :: MonadImport m => TxData -> Maybe BlockRef -> m ()
 confTx t mbr = do
     replaceAddressTx t new
     forM_ (zip [0 ..] (txOut (txData t))) $ \(n, o) -> do
         let op = OutPoint (txHash (txData t)) n
         adjustAddressOutput op o old new
+    rbf <- isRBF new (txData t)
+    let td = t { txDataBlock = new, txDataRBF = rbf }
     insertTx td
     updateMempool td
   where
     new = fromMaybe (MemRef (txDataTime t)) mbr
     old = txDataBlock t
-    td = t {txDataBlock = new}
 
 freeOutputs
     :: MonadImport m
     => Bool -- ^ only delete transaction if unconfirmed
     -> Bool -- ^ only delete RBF
-    -> [Tx]
-    -> WriterT m ()
-freeOutputs memonly rbfcheck txs =
-    forM_ txs $ \tx ->
+    -> Tx
+    -> m ()
+freeOutputs memonly rbfcheck tx =
     forM_ (prevOuts tx) $ \op ->
-    unless (outPointHash op `S.member` ths) $
-    loadPrevOut op >>
-    getUnspent op >>= \case
-        Just _ -> return ()
-        Nothing -> getSpender op >>= \case
-            Just Spender { spenderHash = h }
-                | h == txHash tx -> return ()
-                | otherwise -> deleteTx memonly rbfcheck h
-            Nothing -> do
-                $(logErrorS) "BlockStore" $
-                    "Missing unspent output for tx: "
-                    <> txHashToHex (txHash tx)
-                throwError Orphan
-  where
-    ths = S.fromList (map txHash txs)
+    getUnspent op >>= \u -> when (isNothing u) $
+    getSpender op >>= \p -> forM_ p $ \s ->
+    unless (spenderHash s == txHash tx) $
+    deleteTx memonly rbfcheck (spenderHash s)
 
-deleteConfirmedTx :: MonadImport m => TxHash -> WriterT m ()
+deleteConfirmedTx :: MonadImport m => TxHash -> m ()
 deleteConfirmedTx = deleteTx False False
 
-deleteUnconfirmedTx :: MonadImport m => Bool -> TxHash -> WriterT m ()
+deleteUnconfirmedTx :: MonadImport m => Bool -> TxHash -> m ()
 deleteUnconfirmedTx rbfcheck th =
     getActiveTxData th >>= \case
-        Just _ ->
-            deleteTx True rbfcheck th
-        Nothing ->
-            $(logDebugS) "BlockStore" $
-            "Not found or already deleted: " <> txHashToHex th
+        Just _ -> deleteTx True rbfcheck th
+        Nothing -> $(logDebugS) "BlockStore" $
+                   "Not found or already deleted: " <> txHashToHex th
 
-deleteTx
-    :: MonadImport m
-    => Bool -- ^ only delete transaction if unconfirmed
-    -> Bool -- ^ only delete RBF
-    -> TxHash
-    -> WriterT m ()
+deleteTx :: MonadImport m
+         => Bool -- ^ only delete transaction if unconfirmed
+         -> Bool -- ^ only delete RBF
+         -> TxHash
+         -> m ()
 deleteTx memonly rbfcheck th =
     getChain memonly rbfcheck th >>=
     mapM_ (deleteSingleTx . txHash)
 
-getChain
-    :: MonadImport m
-    => Bool -- ^ only delete transaction if unconfirmed
-    -> Bool -- ^ only delete RBF
-    -> TxHash
-    -> WriterT m [Tx]
+getChain :: MonadImport m
+         => Bool -- ^ only delete transaction if unconfirmed
+         -> Bool -- ^ only delete RBF
+         -> TxHash
+         -> m [Tx]
 getChain memonly rbfcheck th =
     fmap sort_clean $
     getActiveTxData th >>= \case
@@ -484,30 +415,29 @@ getChain memonly rbfcheck th =
                 "Transaction not found: " <> txHashToHex th
             throwError TxNotFound
         Just td
-            | memonly && confirmed (txDataBlock td) -> do
-                  $(logErrorS) "BlockStore" $
-                      "Transaction already confirmed: "
-                      <> txHashToHex th
-                  throwError TxConfirmed
-            | rbfcheck ->
-                isRBF (txDataBlock td) (txData td) >>= \case
-                    True -> go td
-                    False -> do
-                        $(logErrorS) "BlockStore" $
-                            "Double-spending transaction: "
-                            <> txHashToHex th
-                        throwError DoubleSpend
-            | otherwise -> go td
+          | memonly && confirmed (txDataBlock td) -> do
+            $(logErrorS) "BlockStore" $
+                "Transaction already confirmed: "
+                <> txHashToHex th
+            throwError TxConfirmed
+          | rbfcheck ->
+            isRBF (txDataBlock td) (txData td) >>= \case
+                True -> go td
+                False -> do
+                    $(logErrorS) "BlockStore" $
+                        "Double-spending transaction: "
+                        <> txHashToHex th
+                    throwError DoubleSpend
+          | otherwise -> go td
   where
-    sort_clean =
-        reverse . map snd . sortTxs . nub'
+    sort_clean = reverse . map snd . sortTxs . nub'
     go td = do
         let tx = txData td
         ss <- nub' . map spenderHash . I.elems <$> getSpenders th
         xs <- concat <$> mapM (getChain memonly rbfcheck) ss
         return $ tx : xs
 
-deleteSingleTx :: MonadImport m => TxHash -> WriterT m ()
+deleteSingleTx :: MonadImport m => TxHash -> m ()
 deleteSingleTx th =
     getActiveTxData th >>= \case
         Nothing -> do
@@ -518,99 +448,93 @@ deleteSingleTx th =
             $(logDebugS) "BlockStore" $
                 "Deleting tx: " <> txHashToHex th
             getSpenders th >>= \case
-                m | I.null m -> do
-                        preLoadMemory [txData td]
-                        runTx $ commitDelTx td
+                m | I.null m -> commitDelTx td
                   | otherwise -> do
                         $(logErrorS) "BlockStore" $
                             "Tried to delete spent tx: "
                             <> txHashToHex th
                         throwError TxSpent
 
-commitDelTx :: TxData -> MemoryTx ()
-commitDelTx = commitModTx False []
+commitDelTx :: MonadImport m => TxData -> m ()
+commitDelTx = commitModTx False
 
-commitAddTx :: [Unspent] -> TxData -> MemoryTx ()
+commitAddTx :: MonadImport m => TxData -> m ()
 commitAddTx = commitModTx True
 
-commitModTx :: Bool -> [Unspent] -> TxData -> MemoryTx ()
-commitModTx add us td = do
+commitModTx :: MonadImport m => Bool -> TxData -> m ()
+commitModTx add tx_data = do
     mapM_ mod_addr_tx (txDataAddresses td)
     mod_outputs
     mod_unspent
-    insertTx td'
-    updateMempool td'
+    insertTx td
+    updateMempool td
   where
-    td' = td { txDataDeleted = not add }
-    tx_ref = TxRef (txDataBlock td) (txHash (txData td))
-    mod_addr_tx a | add = do
-                        insertAddrTx a tx_ref
-                        modAddressCount add a
-                  | otherwise = do
-                        deleteAddrTx a tx_ref
-                        modAddressCount add a
-    mod_unspent | add = spendOutputs us td
-                | otherwise = unspendOutputs td
-    mod_outputs | add = addOutputs td
-                | otherwise = delOutputs td
+    tx = txData td
+    br = txDataBlock td
+    td = tx_data { txDataDeleted = not add }
+    tx_ref = TxRef br (txHash tx)
+    mod_addr_tx a
+      | add = do
+            insertAddrTx a tx_ref
+            modAddressCount add a
+      | otherwise = do
+            deleteAddrTx a tx_ref
+            modAddressCount add a
+    mod_unspent
+      | add = spendOutputs tx
+      | otherwise = unspendOutputs tx
+    mod_outputs
+      | add = addOutputs br tx
+      | otherwise = delOutputs br tx
 
-updateMempool :: TxData -> MemoryTx ()
-updateMempool td = do
-    mp <- getMempool
-    setMempool (f mp)
-  where
-    f mp | txDataDeleted td || confirmed (txDataBlock td) =
-           filter ((/= txHash (txData td)) . txRefHash) mp
-         | otherwise =
-           sortOn Down $ TxRef (txDataBlock td) (txHash (txData td)) : mp
+updateMempool :: MonadImport m => TxData -> m ()
+updateMempool td@TxData{txDataDeleted = True} =
+    deleteFromMempool (txHash (txData td))
+updateMempool td@TxData{txDataDeleted = False, txDataBlock = MemRef t} =
+    addToMempool (txHash (txData td)) t
+updateMempool td@TxData{txDataBlock = BlockRef{}} =
+    deleteFromMempool (txHash (txData td))
 
-spendOutputs :: [Unspent] -> TxData -> MemoryTx ()
-spendOutputs us td =
-    zipWithM_ (spendOutput (txHash (txData td))) [0 ..] us
+spendOutputs :: MonadImport m => Tx -> m ()
+spendOutputs tx =
+    zipWithM_ (spendOutput (txHash tx)) [0 ..] (prevOuts tx)
 
-addOutputs :: TxData -> MemoryTx ()
-addOutputs td =
-    zipWithM_
-        (addOutput (txDataBlock td) . OutPoint (txHash (txData td)))
-        [0 ..]
-        (txOut (txData td))
+addOutputs :: MonadImport m => BlockRef -> Tx -> m ()
+addOutputs br tx =
+    zipWithM_ (addOutput br . OutPoint (txHash tx)) [0 ..] (txOut tx)
 
 isRBF :: StoreReadBase m
       => BlockRef
       -> Tx
       -> m Bool
 isRBF br tx
-    | confirmed br =
-          return False
-    | otherwise =
-          getNetwork >>= \net ->
-              if getReplaceByFee net
-              then go
-              else return False
+  | confirmed br = return False
+  | otherwise =
+    getNetwork >>= \net ->
+        if getReplaceByFee net
+        then go
+        else return False
   where
-    go | any ((< 0xffffffff - 1) . txInSequence) (txIn tx) =
-             return True
-       | otherwise =
-             carry_on
+    go | any ((< 0xffffffff - 1) . txInSequence) (txIn tx) = return True
+       | otherwise = carry_on
     carry_on =
         let hs = nub' $ map (outPointHash . prevOutput) (txIn tx)
             ck [] = return False
             ck (h : hs') =
                 getActiveTxData h >>= \case
                     Nothing -> return False
-                    Just t
-                        | confirmed (txDataBlock t) -> ck hs'
-                        | txDataRBF t -> return True
-                        | otherwise -> ck hs'
+                    Just t | confirmed (txDataBlock t) -> ck hs'
+                           | txDataRBF t -> return True
+                           | otherwise -> ck hs'
          in ck hs
 
-addOutput :: BlockRef -> OutPoint -> TxOut -> MemoryTx ()
+addOutput :: MonadImport m => BlockRef -> OutPoint -> TxOut -> m ()
 addOutput = modOutput True
 
-delOutput :: BlockRef -> OutPoint -> TxOut -> MemoryTx ()
+delOutput :: MonadImport m => BlockRef -> OutPoint -> TxOut -> m ()
 delOutput = modOutput False
 
-modOutput :: Bool -> BlockRef -> OutPoint -> TxOut -> MemoryTx ()
+modOutput :: MonadImport m => Bool -> BlockRef -> OutPoint -> TxOut -> m ()
 modOutput add br op o = do
     mod_unspent
     forM_ ma $ \a -> do
@@ -632,15 +556,13 @@ modOutput add br op o = do
     mod_addr_unspent | add = insertAddrUnspent
                      | otherwise = deleteAddrUnspent
 
-delOutputs :: TxData -> MemoryTx ()
-delOutputs td =
-    forM_ (zip [0..] outs) $ \(i, o) -> do
-        let op = OutPoint (txHash (txData td)) i
-        delOutput (txDataBlock td) op o
-  where
-    outs = txOut (txData td)
+delOutputs :: MonadImport m => BlockRef -> Tx -> m ()
+delOutputs br tx =
+    forM_ (zip [0..] (txOut tx)) $ \(i, o) -> do
+        let op = OutPoint (txHash tx) i
+        delOutput br op o
 
-getImportTxData :: MonadImport m => TxHash -> WriterT m TxData
+getImportTxData :: MonadImport m => TxHash -> m TxData
 getImportTxData th =
     getActiveTxData th >>= \case
         Nothing -> do
@@ -653,34 +575,33 @@ getTxOut i tx = do
     guard (fromIntegral i < length (txOut tx))
     return $ txOut tx !! fromIntegral i
 
-spendOutput :: TxHash -> Word32 -> Unspent -> MemoryTx ()
-spendOutput th ix u = do
-    insertSpender (unspentPoint u) (Spender th ix)
+spendOutput :: MonadImport m => TxHash -> Word32 -> OutPoint -> m ()
+spendOutput th ix op = do
+    u <- getUnspent op >>= \case
+        Just u -> return u
+        Nothing -> error $ "Could not find UTXO to spend: " <> show op
+    deleteUnspent op
+    insertSpender op (Spender th ix)
     let pk = B.Short.fromShort (unspentScript u)
-    case scriptToAddressBS pk of
-        Left _ -> return ()
-        Right a -> do
-            decreaseBalance
-                (confirmed (unspentBlock u))
-                a
-                (unspentAmount u)
-            deleteAddrUnspent a u
-    deleteUnspent (unspentPoint u)
+    forM_ (scriptToAddressBS pk) $ \a -> do
+        decreaseBalance
+            (confirmed (unspentBlock u))
+            a
+            (unspentAmount u)
+        deleteAddrUnspent a u
 
-unspendOutputs :: TxData -> MemoryTx ()
-unspendOutputs td = mapM_ unspendOutput (prevOuts (txData td))
+unspendOutputs :: MonadImport m => Tx -> m ()
+unspendOutputs = mapM_ unspendOutput . prevOuts
 
-unspendOutput :: OutPoint -> MemoryTx ()
+unspendOutput :: MonadImport m => OutPoint -> m ()
 unspendOutput op = do
     t <- getActiveTxData (outPointHash op) >>= \case
-        Nothing ->
-            error $ "Could not find tx data: " <> show (outPointHash op)
+        Nothing -> error $ "Could not find tx data: " <> show (outPointHash op)
         Just t -> return t
-    o <- case getTxOut (outPointIndex op) (txData t) of
-        Nothing ->
-            error $ "Could not find output: " <> show op
-        Just o -> return o
-    let m = eitherToMaybe (scriptToAddressBS (scriptOutput o))
+    let o = fromMaybe
+            (error ("Could not find output: " <> show op))
+            (getTxOut (outPointIndex op) (txData t))
+        m = eitherToMaybe (scriptToAddressBS (scriptOutput o))
         u = Unspent { unspentAmount = outValue o
                     , unspentBlock = txDataBlock t
                     , unspentScript = B.Short.toShort (scriptOutput o)
@@ -693,23 +614,23 @@ unspendOutput op = do
         insertAddrUnspent a u
         increaseBalance (confirmed (unspentBlock u)) a (outValue o)
 
-modifyReceived :: Address -> (Word64 -> Word64) -> MemoryTx ()
+modifyReceived :: MonadImport m => Address -> (Word64 -> Word64) -> m ()
 modifyReceived a f = do
     b <- getDefaultBalance a
     setBalance b { balanceTotalReceived = f (balanceTotalReceived b) }
 
-decreaseBalance :: Bool -> Address -> Word64 -> MemoryTx ()
+decreaseBalance :: MonadImport m => Bool -> Address -> Word64 -> m ()
 decreaseBalance conf = modBalance conf False
 
-increaseBalance :: Bool -> Address -> Word64 -> MemoryTx ()
+increaseBalance :: MonadImport m => Bool -> Address -> Word64 -> m ()
 increaseBalance conf = modBalance conf True
 
-modBalance
-    :: Bool -- ^ confirmed
-    -> Bool -- ^ add
-    -> Address
-    -> Word64
-    -> MemoryTx ()
+modBalance :: MonadImport m
+           => Bool -- ^ confirmed
+           -> Bool -- ^ add
+           -> Address
+           -> Word64
+           -> m ()
 modBalance conf add a val = do
     b <- getDefaultBalance a
     setBalance $ (g . f) b
@@ -720,7 +641,7 @@ modBalance conf add a val = do
     m | add = (+)
       | otherwise = subtract
 
-modAddressCount :: Bool -> Address -> MemoryTx ()
+modAddressCount :: MonadImport m => Bool -> Address -> m ()
 modAddressCount add a = do
     b <- getDefaultBalance a
     setBalance b {balanceTxCount = f (balanceTxCount b)}
@@ -749,12 +670,3 @@ prevOuts tx = filter (/= nullOutPoint) (map prevOutput (txIn tx))
 
 testPresent :: StoreReadBase m => Tx -> m Bool
 testPresent tx = isJust <$> getActiveTxData (txHash tx)
-
-logOutput :: ( ConvertibleStrings Text a
-             , ConvertibleStrings String a
-             , Semigroup a
-             , IsString a
-             )
-          => OutPoint
-          -> a
-logOutput (OutPoint h i) = cs (txHashToHex h) <> ":" <> cs (show i)
