@@ -7,6 +7,7 @@ module Haskoin.Store.Manager
 
 import           Control.Monad                 (forever, unless, when)
 import           Control.Monad.Logger          (MonadLoggerIO)
+import           Control.Monad.Reader          (ReaderT (ReaderT), runReaderT)
 import           Data.Serialize                (decode)
 import           Data.Time.Clock               (NominalDiffTime)
 import           Data.Word                     (Word32)
@@ -26,30 +27,28 @@ import           Haskoin.Node                  (Chain, ChainEvent (..),
                                                 WithConnection, withNode)
 import           Haskoin.Store.BlockStore      (BlockStore,
                                                 BlockStoreConfig (..),
-                                                blockStore, blockStoreBlockSTM,
+                                                blockStoreBlockSTM,
                                                 blockStoreHeadSTM,
                                                 blockStoreNotFoundSTM,
                                                 blockStorePeerConnectSTM,
                                                 blockStorePeerDisconnectSTM,
                                                 blockStoreTxHashSTM,
-                                                blockStoreTxSTM)
+                                                blockStoreTxSTM, withBlockStore)
 import           Haskoin.Store.Cache           (CacheConfig (..), CacheWriter,
-                                                cacheNewBlock, cacheNewTx,
-                                                cachePing, cacheWriter,
-                                                connectRedis)
+                                                cacheNewBlock, cacheWriter,
+                                                connectRedis, newCacheMetrics)
 import           Haskoin.Store.Common          (StoreEvent (..))
 import           Haskoin.Store.Database.Reader (DatabaseReader (..),
-                                                connectRocksDB,
+                                                DatabaseReaderT,
                                                 withDatabaseReader)
 import           Network.Socket                (SockAddr (..))
 import           NQE                           (Inbox, Process (..), Publisher,
-                                                newMailbox, publishSTM, receive,
+                                                publishSTM, receive,
                                                 withProcess, withPublisher,
                                                 withSubscription)
-import           System.Random                 (randomRIO)
+import qualified System.Metrics                as Metrics (Store)
 import           UnliftIO                      (MonadIO, MonadUnliftIO, STM,
-                                                atomically, liftIO, link,
-                                                withAsync)
+                                                atomically, link, withAsync)
 import           UnliftIO.Concurrent           (threadDelay)
 
 -- | Store mailboxes.
@@ -67,49 +66,55 @@ data Store =
 -- | Configuration for a 'Store'.
 data StoreConfig =
     StoreConfig
-        { storeConfMaxPeers    :: !Int
+        { storeConfMaxPeers        :: !Int
       -- ^ max peers to connect to
-        , storeConfInitPeers   :: ![HostPort]
+        , storeConfInitPeers       :: ![HostPort]
       -- ^ static set of peers to connect to
-        , storeConfDiscover    :: !Bool
+        , storeConfDiscover        :: !Bool
       -- ^ discover new peers
-        , storeConfDB          :: !FilePath
+        , storeConfDB              :: !FilePath
       -- ^ RocksDB database path
-        , storeConfNetwork     :: !Network
+        , storeConfNetwork         :: !Network
       -- ^ network constants
-        , storeConfCache       :: !(Maybe String)
+        , storeConfCache           :: !(Maybe String)
       -- ^ Redis cache configuration
-        , storeConfInitialGap  :: !Word32
+        , storeConfInitialGap      :: !Word32
       -- ^ gap on extended public key with no transactions
-        , storeConfGap         :: !Word32
+        , storeConfGap             :: !Word32
       -- ^ gap for extended public keys
-        , storeConfCacheMin    :: !Int
+        , storeConfCacheMin        :: !Int
       -- ^ cache xpubs with more than this many used addresses
-        , storeConfMaxKeys     :: !Integer
+        , storeConfMaxKeys         :: !Integer
       -- ^ maximum number of keys in Redis cache
-        , storeConfWipeMempool :: !Bool
+        , storeConfNoMempool       :: !Bool
+      -- ^ do not index new mempool transactions
+        , storeConfWipeMempool     :: !Bool
       -- ^ wipe mempool when starting
-        , storeConfPeerTimeout :: !NominalDiffTime
+        , storeConfPeerTimeout     :: !NominalDiffTime
       -- ^ disconnect peer if message not received for this many seconds
-        , storeConfPeerMaxLife :: !NominalDiffTime
+        , storeConfPeerMaxLife     :: !NominalDiffTime
       -- ^ disconnect peer if it has been connected this long
-        , storeConfConnect     :: !(SockAddr -> WithConnection)
+        , storeConfConnect         :: !(SockAddr -> WithConnection)
       -- ^ connect to peers using the function 'withConnection'
+        , storeConfCacheRefresh    :: !Int
+      -- ^ refresh the cache this often (milliseconds)
+        , storeConfCacheRetryDelay :: !Int
+      -- ^ delay in microseconds to retry getting cache lock
+        , storeConfStats           :: !(Maybe Metrics.Store)
+      -- ^ stats store
         }
 
 withStore :: (MonadLoggerIO m, MonadUnliftIO m)
           => StoreConfig -> (Store -> m a) -> m a
 withStore cfg action =
-    connectDB cfg >>= \db ->
-    newMailbox >>= \(bi, b) ->
+    connectDB cfg $ ReaderT $ \db ->
     withPublisher $ \pub ->
     withPublisher $ \node_pub ->
     withSubscription node_pub $ \node_sub ->
     withNode (nodeCfg cfg db node_pub) $ \node ->
     withCache cfg (nodeChain node) db pub $ \mcache ->
-    withAsync (nodeForwarder b pub node_sub) $ \a1 ->
-    withAsync (blockStore (blockStoreCfg cfg node pub db) bi) $ \a2 ->
-    link a1 >> link a2 >>
+    withBlockStore (blockStoreCfg cfg node pub db) $ \b ->
+    withAsync (nodeForwarder b pub node_sub) $ \a1 -> link a1 >>
     action Store { storeManager = nodeManager node
                  , storeChain = nodeChain node
                  , storeBlock = b
@@ -119,9 +124,9 @@ withStore cfg action =
                  , storeNetwork = storeConfNetwork cfg
                  }
 
-connectDB :: MonadIO m => StoreConfig -> m DatabaseReader
+connectDB :: MonadUnliftIO m => StoreConfig -> DatabaseReaderT m a -> m a
 connectDB cfg =
-    connectRocksDB
+    withDatabaseReader
           (storeConfNetwork cfg)
           (storeConfInitialGap cfg)
           (storeConfGap cfg)
@@ -139,8 +144,10 @@ blockStoreCfg cfg node pub db =
         , blockConfListener = pub
         , blockConfDB = db
         , blockConfNet = storeConfNetwork cfg
+        , blockConfNoMempool = storeConfNoMempool cfg
         , blockConfWipeMempool = storeConfWipeMempool cfg
         , blockConfPeerTimeout = storeConfPeerTimeout cfg
+        , blockConfStats = storeConfStats cfg
         }
 
 nodeCfg :: StoreConfig
@@ -151,6 +158,7 @@ nodeCfg cfg db pub =
     NodeConfig
         { nodeConfMaxPeers = storeConfMaxPeers cfg
         , nodeConfDB = databaseHandle db
+        , nodeConfColumnFamily = Nothing
         , nodeConfPeers = storeConfInitPeers cfg
         , nodeConfDiscover = storeConfDiscover cfg
         , nodeConfEvents = pub
@@ -176,38 +184,37 @@ withCache cfg chain db pub action =
         Nothing ->
             action Nothing
         Just redisurl ->
+            mapM newCacheMetrics (storeConfStats cfg) >>= \metrics ->
             connectRedis redisurl >>= \conn ->
             withSubscription pub $ \evts ->
-            withProcess (f conn) $ \p ->
-            cacheWriterProcesses evts (getProcessMailbox p) $
-            action (Just (c conn))
+            withProcess (f conn metrics) $ \p ->
+            cacheWriterProcesses crefresh evts (getProcessMailbox p) $
+            action (Just (c conn metrics))
   where
-    f conn cwinbox =
-        withDatabaseReader db $
-        cacheWriter (c conn) cwinbox
-    c conn = CacheConfig
-                { cacheConn = conn
-                , cacheMin = storeConfCacheMin cfg
-                , cacheChain = chain
-                , cacheMax = storeConfMaxKeys cfg
-                }
+    crefresh = storeConfCacheRefresh cfg
+    f conn metrics cwinbox =
+        runReaderT (cacheWriter (c conn metrics) cwinbox) db
+    c conn metrics =
+        CacheConfig
+           { cacheConn = conn
+           , cacheMin = storeConfCacheMin cfg
+           , cacheChain = chain
+           , cacheMax = storeConfMaxKeys cfg
+           , cacheRefresh = storeConfCacheRefresh cfg
+           , cacheRetryDelay = storeConfCacheRetryDelay cfg
+           , cacheMetrics = metrics
+           }
 
 cacheWriterProcesses :: MonadUnliftIO m
-                     => Inbox StoreEvent
+                     => Int
+                     -> Inbox StoreEvent
                      -> CacheWriter
                      -> m a
                      -> m a
-cacheWriterProcesses evts cwm action =
-    withAsync events $ \a1 ->
-    withAsync ping   $ \a2 ->
-    link a1 >> link a2 >> action
+cacheWriterProcesses crefresh evts cwm action =
+    withAsync events $ \a1 -> link a1 >> action
   where
     events = cacheWriterEvents evts cwm
-    ping = forever $ do
-        time <- liftIO $ randomRIO (5 * second, 15 * second)
-        threadDelay time
-        cachePing cwm
-    second = 1000000
 
 cacheWriterEvents :: MonadIO m => Inbox StoreEvent -> CacheWriter -> m ()
 cacheWriterEvents evts cwm =
@@ -216,9 +223,8 @@ cacheWriterEvents evts cwm =
     e `cacheWriterDispatch` cwm
 
 cacheWriterDispatch :: MonadIO m => StoreEvent -> CacheWriter -> m ()
-cacheWriterDispatch (StoreBestBlock _)   = cacheNewBlock
-cacheWriterDispatch (StoreMempoolNew th) = cacheNewTx th
-cacheWriterDispatch _                    = const (return ())
+cacheWriterDispatch (StoreBestBlock _) = cacheNewBlock
+cacheWriterDispatch _                  = const (return ())
 
 nodeForwarder :: MonadIO m
               => BlockStore
