@@ -7,6 +7,7 @@ module Haskoin.StoreSpec (spec) where
 import           Conduit
 import           Control.Monad
 import           Control.Monad.Logger
+import           Control.Monad.Reader
 import           Data.ByteString        (ByteString)
 import qualified Data.ByteString        as B
 import           Data.ByteString.Base64
@@ -19,10 +20,13 @@ import           Data.Word
 import           Haskoin
 import           Haskoin.Node
 import           Haskoin.Store
+import           Haskoin.Util.Arbitrary
 import           Network.Socket
 import           NQE
 import           System.Random
 import           Test.Hspec
+import           Test.Hspec.QuickCheck
+import           Test.QuickCheck
 import           UnliftIO
 
 data TestStore = TestStore
@@ -33,7 +37,8 @@ data TestStore = TestStore
     }
 
 spec :: Spec
-spec = describe "Download" $ do
+spec = do
+  describe "Download" $ do
     it "gets 8 blocks" $
         withTestStore bchRegTest "eight-blocks" $ \TestStore {..} -> do
         bs <- replicateM 8 . receiveMatch testStoreEvents $ \case
@@ -47,7 +52,7 @@ spec = describe "Download" $ do
         bestHeight `shouldBe` 8
     it "get a block and its transactions" $
         withTestStore bchRegTest "get-block-txs" $ \TestStore {..} ->
-        withDatabaseReader testStoreDB $ do
+        flip runReaderT testStoreDB $ do
         let h1 = "5369ef2386c72acdf513ffd80aeba2a1774e2f004d120761e54a8bf614173f3e"
             get_the_block h =
                 receive testStoreEvents >>= \case
@@ -86,10 +91,14 @@ withTestStore net t f =
               , storeConfInitialGap = 20
               , storeConfCacheMin = 100
               , storeConfMaxKeys = 100 * 1000 * 1000
+              , storeConfNoMempool = False
               , storeConfWipeMempool = False
               , storeConfPeerTimeout = 60
               , storeConfPeerMaxLife = 48 * 3600
               , storeConfConnect = dummyPeerConnect net ad
+              , storeConfCacheRefresh = 750
+              , storeConfCacheRetryDelay = 100000
+              , storeConfStats = Nothing
               }
     withStore cfg $ \Store {..} ->
         withSubscription storePublisher $ \sub ->
